;; Copyright (c) 2022 Sean Cole <scole@NetBSD.org>
;; All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions
;; are met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;;
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
;; TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
;; PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
;; BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
;; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
;; SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
;; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
;; CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
;; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
;; POSSIBILITY OF SUCH DAMAGE.

;; utilities for working with thai text in emacs buffers.  includes
;; functions to split/unsplit thai strings, thai time conversion,
;; download and play mp3 audio for thai words from thai-language.com,
;; clickthai-online.com, thai2english.com, and lingopolo.org, and currency
;; conversions from x-rates.com.  Also, it can look up thai words if
;; vocabulary files are created.  The 'customize interface is available
;; for some settings.
;;
;; HOW TO USE
;; - in ~/.emacs/init.el, add library path and a require, e.g.:
;;     (add-to-list 'load-path "ST_SHARE_DIR")
;;     (require 'pthai)
;;   or load module directly:
;;     (load-file "ST_SHARE_DIR/pthai.el")
;;
;; - there is a 'pthai-mode (M-x pthai-mode) that remaps some common
;;   word key-bindings and binds many commonly used pthai functions.
;;   see define-minor-mode 'pthai-mode in this file for keybindings
;;
;; - or you can override keys manually in ~/.emacs/init.el,
;;    for example:
;;    (global-set-key [f7] 'pthai-lookup-at-point)
;;    (global-set-key [f8] 'pthai-say-word-at-point)
;;    (global-set-key [f9] 'pthai-say-string-at-point)
;;    (global-set-key [f10] 'pthai-say-line)
;;    (global-set-key [s-right] 'forward-thaiword)
;;    (global-set-key [s-left] 'backward-thaiword)
;;
;; - 'customize should be used to add paths to dictionaries and
;;   to adjust other settings
;; - M-x apropos pthai for available functions
;;
;; TODO
;; - play audio for individual vowels
;; - when parsing thai words, need separate some thai punctuation like "พีฺฺฺ่"
;; - interface with pkgsrc/textproc/translate-shell, licensing (?)
;; - handle ๆ/" better, repeating last word and word breaking (?)
;; - spell replace word ถุกๆ => ถูกๆๆ
;; - create an info file for this module
;; - byte compile module
;; - when doing pthai-lookup, add "classifier" to definition if word classifier (?)
;; - fix pthai-lookup to return classifier info if word only has class info
;; - loading module 2x messes up word counts/defvar problem with 'pthai-dictionary-list?
;; - better spelling suggestions with word distributions
;; - get previous thai word when not on space only (ไก่) vs ไก่
;;   by adding function to check point/previous point, shift back 1 if needed
;; - add menu in pthai-mode (?)
;; - look up word definitions on the fly
;; - say and look up consonants, but don't use for spell checking
;; - don't spell check single word consonants, eg, "ก:"
;; - refresh modeline col/row positions when saying words and/or progress reporting
;; - add *sounds-like function to lookup similar sounding words
;; - elisp/pthai-rwb splitter to handle long string & misspellings
;; - add hook to override vocabulary words
;; - move to end of word after *insert *complete-word functions
;; - pause 1 sec when reading word with no audio/zero-length audio,
;;   in pthai-download-and-play? eg. ชอน/มิชชัน/คาร์เตอร์
;; - words with only classifier definition not found with pthai-unknowns(?)
;; - tabulate downloaded word counts per letter
;; - issues with sox under cygwin like utf-8 filenames and audio distorting(?)
;;
;; known issues:
;; - when linting or compiling file, need to "(require 'seq)" first?
;;
(require 'cus-edit)   ;; custom* customize*
(require 'easy-mmode) ;; define-minor-mode
(require 'mule-util)  ;; nested alist functions
(require 'pulse)      ;; pulse* functions
(require 'seq)        ;; seq-* functions
(require 'simple)     ;; count-* functions
(require 'sort)       ;; sort-lines
(require 'subr-x)     ;; string-trim* functions
(require 'thai-word)  ;; thai-word-table
(require 'thingatpt)  ;; thing-at-point*
(require 'url)        ;; url-* functions

;; xxx "special" vars to set before loading module, not the emacs way?
(unless (boundp 'pthai-bootstrap)
  (defvar pthai-bootstrap nil "nil unless building for pkgsrc"))

(unless (boundp 'pthai-verbose-wordloads)
  (defvar pthai-verbose-wordloads t
    "if non-nil, display word counts when loading dictionaries"))

(defconst pthai-pkg-wordlist
  "ST_SHARE_DIR/words"
  "path to pkg wordlist file")

(defconst pthai-pkg-sample-dictionary
  "ST_SHARE_DIR/sampledict.txt"
  "path to pkg sample dictionary file")
  
(defvar pthai-default-directory (concat user-emacs-directory "pthai/")
  "default pthai directory (ensure ends with directory separator)")

(defvar pthai-dictionary (make-hash-table :test 'equal)
   "hash of thai dictionary words.  each word has a dictionary entry, \n\
the form of the dictionary entry is:\n\
thaiword => '( definition english_classifier thai_classisifier)\n\ 
which are all lists of strings.  empty definitions should be defined as nil:\n\
\"ไก่\" => ( '(\"chicken\") nil '(\"ตัว\") ) \n\n\
if a thai word is known but no definition has been created \n\
 '( nil nil nil ) \nwill be its entry ")

(defconst pthai-consonants
  (split-string "กขฃคตฆงจฉชซณญฎฏฑฒณฐดตถทนบธปผฝพฟภมยรลวศษสหฬอฮ" "" t)
  "list of thai consonants")

;; https://www.omniglot.com/writing/thai.htm
(defconst pthai-vowels
  (list "อะ" "อั" "อา" "อิ" "อี" "อึ" "อื"
	"อีอ" "อุ" "อู" "เอะ" "เอ็" "เอ" "แอะ"
	"แอ็" "แอ" "โอะ" "โอ" "เอาะ" "อ็อ" "ออ"
	"อ็" "เออะ" "เออ" "เอิ"
	"เอียะ" "เอีย" "เอือะ" "เอือ" "อัวะ" "อัว" "อว"
	"อิว" "เอ็อ" "เอว" "แอว" "เอา" "อาว" "เอียว"
	"อัย" "ใอ" "ไอ" "ไอย" "อาย" "อ็อย" "ออย"
	"โอย" "อุย" "เอย" "อวย" "เอือย" "อำ" "ฤ"
	"ฤๅ" "ฦ" "ฦๅ" "อํ" "อ์")
  "list of thai vowels (replace first 'อ' with consonant)")

(defconst pthai-numbers
  (split-string "๐๑๒๓๔๕๖๗๘๙" "" t)
  "list of thai numbers")

(defconst pthai-regexp-single-consonant
  "^[\u0e01-\u0e2e]\\{1\\}$"
  "single unicode thai consonant [ก ข ฃ ... อ ฮ]")

(defconst pthai-regexp-thai-number
  "^[\u0e50-\u0e59]+$"
  "thai unicode number [๐๑๒๓๔๕๖๗๘๙]")

(defconst pthai-regexp-any-number
  "^\\([\u0e50-\u0e59]+\\|[0-9]+\\)$"
  "number [0-9] or thai unicode number [๐๑๒๓๔๕๖๗๘๙]")

(defconst pthai-regexp-spelling-chars
  "\\([\u0e01-\u0e3a\u0e40-\u0e4f]+\\)"
  "thai unicode characters that should be matched for spelling")

(defconst pthai-regexp-valid-standalone-char
  "\\([\u0e13\u0e2f\u0e3a\u0e3f\u0e46\u0e4d\u0e4f\u0e50-\u0e5b]+\\)"
  "thai unicode characters valid as a stand alone character [ณฯฺ฿ๆํ๏๚๛๐๑๒๓๔๕๖๗๘๙]")

(defvar pthai-rwb-tmp nil "temporary variable for pthai-rwb") 

(defgroup pthai nil
  "Pthai dictionary, word-splitting and audio-playing for Thai words"
  :group 'applications)

(defcustom pthai-split-mode "biggest"
  "Word splitting strategy"
  :group 'pthai
  :set (lambda (sym val) (set-default sym val))
  :type '(radio
	  (const :tag "biggest words possible/fewest words" :value "biggest")
	  (const :tag "smallest words possible/most words" :value "smallest")
	  (const :tag "fewest misspelled words and spelling suggestions" :value "spelling")
	  (const :tag "interactively display choices" :value "interactive")))

(defcustom pthai-splitter-max-recurse-word-length 60
  "maximum string length of thai string used by some recursive
programs when splitting words. if string longer than this value,
don't use recursion for splitting.  set to 0 to disable recursive splitters"
  :group 'pthai
  :type 'integer)

(defcustom pthai-splitter-swath-word-length 80
  "maximum string length of thai string used by swath when
splitting words.  if a string is longer than this setting, don't
use swath for splitting.  set to 0 to disable swath splitting

(xxx swath sometimes hanging for strings longer than ~120 chars ?)"
  :group 'pthai
  :type 'integer)

(defcustom pthai-fetch-clickthai t
  "fetch mp3 audio from clickthai-online.com"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-fetch-lingopolo t
  "fetch mp3 audio from lingopolo.com"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-fetch-thai2english t
  "fetch mp3 audio from thai2english.com"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-fetch-thailanguage t
  "fetch mp3 audio from thai-language.com"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-audio-download t
  "attempt to download audio files for thai words. if disabled,
only already downloaded files will be played"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-audio-download-non-dictionary t
  "attempt to download audio for words not in dictionary"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-audio-display-definition t
  "when playing thai audio also display definition of word if true"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-audio-play t
  "play audio files for thai words. if nil, only the definitions
will be displayed"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-audio-strip-silence t
  "when playing audio words strip the silences from them.  this
speeds up playing and makes playing multiple words sound more
natural"
  :group 'pthai
  :type 'boolean)

(defun pthai-directory-create(dir)
  "prompt user to create a directory if it doesn't exist"
  (unless (file-directory-p dir)
    (if (yes-or-no-p (concat dir " does not exist, create?"))
	(or (make-directory dir)
	    (error "failed to create dir %s" dir)))))

(defcustom pthai-audio-directory
  (concat pthai-default-directory "audio")
  "directory where mp3 audio files are download and saved to, as
well as read from"
  :group 'pthai
  :set (lambda (sym val)
	 (set-default sym val)
	 (unless load-in-progress
	   (pthai-directory-create val)))
  :type 'directory)

(defun pthai-completing-read-full(prompt options &optional default)
  "wrapper for 'completing-read to display full choice list initially"
  (minibuffer-with-setup-hook 'minibuffer-completion-help
    (completing-read prompt options nil nil default)))

(defun pthai-completing-read-dynamic(prompt &optional prefix)
  "wrapper for 'completing-read to dynamically calculate choices
from all words in 'pthai-dictionary starting with opt prefix
(use C-q SPC to enter spaces in minibuffer prompt)"
  (completing-read prompt
		   (completion-table-dynamic
		    (lambda(x) (all-completions x pthai-dictionary)))
		   nil
		   nil
		   prefix))

(defun pthai-nth-delete(n lst)
  "delete nth element from list.  signal error if out of bounds"
  (let ((len (length lst)))
    (cond
     ((or (< n 0) (>= n len))
      (error "out of bounds"))      
     ((= n 0)
      (seq-drop lst 1))
     ((= n (1- len))
      (seq-take lst n))
     (t
      (append (seq-take lst n) (seq-drop lst (1+ n)))))))

(defun pthai-nth-set(elem n lst)
  "set nth element in list to elem.  signal error if out of bounds"
  (let ((len (length lst)))
    (cond
     ((or (< n 0) (>= n len))
      (error "out of bounds"))
     ((= n 0)
      (cons elem (seq-drop lst 1)))
     ((= n (1- len))
      (append (seq-take lst n) (list elem)))
     (t
      (append (seq-take lst n)
	      (list elem)
	      (seq-drop lst (1+ n)))))))

(defun pthai-randomize(lst)
  "take a list and return new list with elements randomized"
  (let (rlst n)
    (while lst
      (setq n (random (length lst))
	    rlst (cons (nth n lst) rlst)
	    lst (pthai-nth-delete n lst)))
    rlst))

;; utilities to manipulate the 'thai-word-table from the 'thai-word
;; library.  these functions could be merged into the 'thai-word
;; module itself, but that module looks to have been written when
;; 'thai-tis620 encoding was commonly used instead of utf-8 for thai
;;
;; 'thai-word-table uses a nested-alist from mule-util which looks
;; like a trie:
;;  '(3585 1                       ;; ก      word   ก
;;    (3591 1                      ;;  ง     word   กง
;;          (3585 t                ;;    ก               
;;                (3634 t          ;;     า              
;;                      (3619 1))));;      ร word   กงการ
;;    (3585 1                      ;;  ก     word   กก 
;;          (3621 1))))            ;;   ล    word   กกล
;;
(defun pthai-twt-words(&optional sorted)
  "return list of all thai words from 'thai-word-table from
'thai-word module.  if option sorted, return them sorted
alphabetically"
  (let ((todo (cdr thai-word-table))
	(words '())
	alist a b c)
    (while todo
      (setq alist (car todo)
	    todo (cdr todo))
	    ;; first of list (3585 1 ...)
      (setq a (nth 0 alist)
	    b (nth 1 alist)
	    c (nthcdr 2 alist))
      ;; convert (3585 1 ...) => ("ก" 1 ...)
      (if (integerp a) (setq a (char-to-string a)))
      ;; word done, but maybe additional merging
      (if (eq b 1) (setq words (cons a words)))
      
      (cond
       ;; already complete word ("กง" 1)
       ((and (stringp a) (not c))
	nil)
       ;; merge like
       ;; ("ก" 1 (3634 t (3621 1 ...))) or
       ;; ("ษั" t ((3625 1 (3600 1)) (3637 1) (3632 1)))
       ((and (stringp a) (consp c))
	(dolist (tmp c)
	  (setq todo (cons
		      (cons
		       (concat a (if (integerp (nth 0 tmp))
				     (char-to-string (nth 0 tmp)) (nth 0 tmp)))
		       (cdr tmp))
		      todo))))
       ;; error
       (t
	(error "parse error reducing"))))
    (if sorted
	(sort words 'string-lessp)
      words)))

(defun pthai-twt-p(thaiword)
 "return t if thaiword is in 'thai-word-table, nil otherwise"
  (let ((result
	 (lookup-nested-alist
	  thaiword thai-word-table (length thaiword) 0 t)))
    (and (listp result)
	 (integerp (car result))
	 (= 1 (car result)))))

(defun pthai-twt-count()
  "return number of words in the 'thai-word-table from the
'thai-word module.  message user when called interactively"
  (interactive)
  (let ((count (length (pthai-twt-words))))
    (when (called-interactively-p 'any)
      (message "%d words" count))
    count))

(defun pthai-twt-read(&optional clear)
  "read 'thai-word-table from 'thai-word into 'pthai-dictionary,
optionally clearing first. return total words in
'pthai-dictionary after loading.  message words in
'thai-word-table and 'pthai-dictionary if called interactively"
  (interactive)
  (let* ((words (pthai-twt-words))
	 (twt-count (length words))
	 (dict-count 0))
    (if clear
	(clrhash pthai-dictionary))
    (dolist (tmp words)
      (if (string-match-p "^\\ct+$" tmp)
     	  (pthai-dictionary-add-word tmp)))
    (setq dict-count (hash-table-count pthai-dictionary))
    (if (called-interactively-p 'any)
	(message "%d words in thai-word-table, %d total words"
		 twt-count dict-count))
    dict-count))

(defun pthai-twt-save(filename)
  "save sorted thai words extracted from 'thai-word-table from
'thai-word module to filename in utf8 format, one word per line.
returns number of words saved"
  (interactive "FName of file to save to: ")
  (let* ((coding-system-for-read 'utf-8)
	 (coding-system-for-write 'utf-8)
	 (buffer-file-coding-system 'utf-8)
	 (words (pthai-twt-words)))
    (with-temp-buffer
      (dolist (tmp words)
	(insert tmp "\n"))
      (sort-lines nil (point-min) (point-max))
      (write-region nil nil filename))
    (length words)))

(defun pthai-twt-update(file &optional append)
  "Update Thai word table by replacing the current word list with
FILE, which is utf-8.  If called with optional append argument,
FILE is appended instead to the current word list.  Does the same
as 'thai-update-word-table, except that function expects
'thai-tis620 encoding which appears to be the default format for
at least emacs version <= 28.0"
  (interactive "FThai word table file: ")
  (let* ((coding-system-for-read 'utf-8)
	 (coding-system-for-write 'utf-8)
	 (buffer-file-coding-system 'utf-8)
	 (temp_file (make-temp-file "pthai-twt-update-620_")))
    (unwind-protect
	(with-temp-buffer
	  (insert-file-contents file)
	  (setq coding-system-for-write 'thai-tis620)
	  (write-file temp_file))
      (thai-update-word-table temp_file append)
      (delete-file temp_file)
      (if pthai-verbose-wordloads
	  (message "%d words in 'thai-word-table" (pthai-twt-count)))
      thai-word-table)))

(defun pthai-rwb(thaistr)
  "return list of space separated string of possible word breaks
for a thai string, uses recursion (recursive word break).  can
only handle known (i.e., in dictionary) thai words.  if no word
breaks found, return '(thaistr)."
  (unless (string-match-p "^\\ct+$" thaistr)
    (error "cannot split non-thai strings: %s" thaistr))
  (setq pthai-rwb-tmp nil)
  (pthai-rwb-helper thaistr (length thaistr) "")
  (cond
   ;; unable to break
   ((not pthai-rwb-tmp)
    (list thaistr))
   ;; strings with no-consecutive/single numbers
   ((not (string-match-p "[\u0e50-\u0e59]\\{2,\\}" thaistr))
    pthai-rwb-tmp)
   ;; strings with >= 2 consecutive digits
   (t
    (seq-uniq
     (mapcar
      (lambda(x)
	(with-temp-buffer
	  (insert x)
	  (goto-char (point-min))
	  (while (re-search-forward
		  "\\([\u0e50-\u0e59]+\\) \\([\u0e50-\u0e59]+\\)" nil t)
	    (replace-match (concat (match-string 1) (match-string 2)))
	    (goto-char (match-beginning 1)))
	  (buffer-string)))
      pthai-rwb-tmp)))))

(defun pthai-rwb-helper(str n result)
  "recursive helper function for pthai-rwb"
  (let ((i 1)
	prefix
	ss)
    (catch 'done
      (while (<= i n)
	(setq prefix (substring-no-properties str 0 i))
	(when (pthai-dictionary-p prefix)
	  (when (= i n)
	    (setq result (concat result prefix)
		  pthai-rwb-tmp (cons result pthai-rwb-tmp))
	    (throw 'done nil))
	  (setq ss (substring-no-properties str i))
	  (pthai-rwb-helper ss (length ss) (concat result prefix " ")))
	(setq i (1+ i))))))

(defun pthai-cycle-split-mode()
  "cycle to next word splitting mode. return string of current
mode.  message user when called interactively"
  (interactive)
  (setq pthai-split-mode (downcase pthai-split-mode))

  (setq pthai-split-mode
	(cond
	 ((string= pthai-split-mode "biggest")
	  "smallest")
	 ((string= pthai-split-mode "smallest")
	  "spelling")
	 ((string= pthai-split-mode "spelling")
	  "interactive")
	 ((string= pthai-split-mode "interactive")
	  "biggest")
	 (t
	  (error "unknown split mode %s" pthai-split-mode))))
  (if (called-interactively-p 'any)
      (message pthai-split-mode))
  pthai-split-mode)

(defun pthai-split-command(command text)
  "run external command to split thai text.  command should be
whitespace separated commandline arguments. returns list of
strings where each string is a group of words separated by
spaces"
  (let ((exe (string-trim (car (split-string command)))))
    (unless (executable-find exe)
      (error "%s not found" exe))
    (split-string
     (shell-command-to-string
      (concat command " \"" text "\"")) "\n" t)))

(defun pthai-split-all(text)
  "split thai string into words. only understands thai character
strings.  returns all found word segmentation combinations as a
list of strings, where each string is a group of words separated
by spaces."
  (let ((len (length text)))
    ;; try to pick fastest split method with most dictionary words. if
    ;; existing words and "short" string, use internal elisp splitter.
    ;; otherwise use external splitter.  swath has more words than
    ;; icu, but sometimes seems to hang for long strings. icu has own
    ;; smaller dictionary but can handle very long strings.
    ;;
    ;; xxx pthai-rwb returns all possible combos, others return "best"
    (cond
     ((and (<= len pthai-splitter-max-recurse-word-length)
	   (pthai-spell-p text))
      (pthai-rwb text))
     ((<= len pthai-splitter-swath-word-length)
      (pthai-split-command "st-swath" text))
     (t
      (pthai-split-command "st-icu" text)))))

(defun pthai-split-best(text &optional smode)
  "split up thai string into words.  words are broken up
depending on 'pthai-split-mode setting and words available in
'pthai-dictionary.  split mode can be specified with optional arg
smode using same values as 'pthai-split-mode setting.  returns a
string containing words separated by spaces"
  (let ((smode (or smode pthai-split-mode))
	slist)
    (cond
     ;; biggest/smallest
     ((or (string= smode "biggest")
	  (string= smode "smallest"))
      (setq slist (sort (pthai-split-all text)
			(lambda (a b) (< (length (split-string a))
    					 (length (split-string b))))))
      (if (string= smode "smallest")
	  (setq slist (nreverse slist)))
      (seq-find (lambda(x) (seq-every-p 'pthai-dictionary-p (split-string x)))
		slist
		(car slist)))
     ;; spelling
     ((string= smode "spelling")
       (setq slist (pthai-spell-suggestions text)
       	     slist (pthai-spell-sort slist))
       (string-join (caar slist) " "))
     ;; interactively
     ((string= smode "interactive")
      (pthai-completing-read-full "split words: " (pthai-split-all text)))
     (t
      (error "invalid split mode (%s)" smode)))))

(defun pthai-bounds-of-thaistring-at-point(&optional cpoint)
  "return beginning and end points of thai string or nil. no word
splitting is done"
  (save-excursion
    (if cpoint 
	(goto-char cpoint))
    ;; on whitespace at end of word, move point back
    (if (and (looking-at-p "[ \f\t\n\r\v]")
	     (string-match-p "\\ct" (char-to-string (preceding-char))))
	(goto-char (1- (point))))
    (if (looking-at-p "\\Ct")
	nil
      (let ((o_pt (point))
	    (p1 (line-beginning-position))
	    (p2 (line-end-position)))
	(if (re-search-backward "\\(\\Ct\\)" p1 t)
	    (setq p1 (match-end 1)))
	(goto-char o_pt)
	(if (re-search-forward "\\(\\ct+\\)" p2 t)
	    (setq p2 (match-end 1)))
	(cons p1 p2)))))

(defun pthai-bounds-offset-of-split-string(s_pt slist)
  "helper function to calculate bounds of contiguous thai string
assumed to be split starting at point s_pt in a buffer. slist is
an array of thai words assumed split correctly"
  (let* ((s_off (list 0))
	 (s_len (mapcar 'length slist)))
    (dolist (tmp (butlast s_len))
      (setq s_off (cons (+ tmp (car s_off)) s_off)))
    (setq s_off
	  (seq-map-indexed
	   (lambda (x n) (cons (+ s_pt x) (+ s_pt x (nth n s_len))))
	   (nreverse s_off)))))

(defun pthai-bounds-of-thaiwords-at-point(&optional cpoint smode)
  "return a list of word bounds for a contiguous thai string at
point. does pthai-word splitting to calculate bounds.  returns
nil if point not at thai string"
  (save-excursion
    (let* ((smode (or smode pthai-split-mode))
	   (bounds (pthai-bounds-of-thaistring-at-point (or cpoint (point))))
	   (p1 (car bounds))
	   (p2 (cdr bounds)))
      (and bounds
	   (pthai-bounds-offset-of-split-string p1
	    (split-string
	     (pthai-split-best (buffer-substring-no-properties p1 p2) smode)))))))

(defun pthai-bounds-of-thaiword-at-point(&optional cpoint smode)
  "Return beginning and end boundaries of a thai word at point.  does
word-splitting to determine the actual thai word.  used with
thingatpoint library to allow 
  (thing-at-point 'thaiword)
  (forward-thaiword)
  (backward-thaiword)
  (beginning-of-thing 'thaiword)
  (end-of-thing 'thaiword)
to work (others?).  returns nil if boundaries not found"
  (save-excursion
    (let* ((smode (or smode pthai-split-mode))
	   (pt (or cpoint (point)))
	   (bounds (pthai-bounds-of-thaiwords-at-point pt smode)))
      (and bounds
	   (seq-find (lambda (x) (and (>= pt (car x)) (< pt (cdr x))))
		     bounds nil)))))

;; add to thingatpoint library
(put 'thaiword 'bounds-of-thing-at-point
     'pthai-bounds-of-thaiword-at-point)

(defun pthai-beginning-of-thaiword()
  "move point to beginning of thaiword"
  (beginning-of-thing 'thaiword))

(defun pthai-end-of-thaiword()
  "move point to end of thaiword"
  (end-of-thing 'thaiword))

(defun forward-thaiword(&optional arg)
  "Move point forward 1 or arg thaiwords (backward if arg is
negative). returns t if moved past a thai word, else nil.
skips over non-thai text.  (see also 'thai-forward-word)"
  (interactive "p")
  (let ((arg (or arg 1))
	(success t)
	tmp)
    ;; backward - move to beginning of closest thaiword backward from point
    (while (and (< arg 0) success)
      (cond
       ((bobp)
	(setq success nil))
       ((looking-at-p "\\ct")
	(setq tmp (car (pthai-bounds-of-thaiword-at-point)))
	(when (= (point) tmp)
	  (setq tmp (1- tmp))
	  (setq arg (1- arg)))
	(goto-char tmp))
       (t
	(unless (and (re-search-backward "\\ct" nil t)
		     (goto-char (car (pthai-bounds-of-thaiword-at-point))))
	  (setq success nil))))
      (setq arg (1+ arg)))
    ;; forward - move to end of closest thai word going right
    (while (and (> arg 0) success)
      (cond
       ((eobp)
	(setq success nil))
       ((looking-at-p "\\ct")
	(goto-char (cdr (pthai-bounds-of-thaiword-at-point))))
       (t
	(unless (and (re-search-forward "\\ct" nil t)
		     (goto-char
		      (cdr (pthai-bounds-of-thaiword-at-point (1- (point))))))
	  (setq success nil))))
      (setq arg (1- arg)))
    success))

(defun backward-thaiword(&optional arg)
  "move backward until encountering the beginning of an thai
  word.  with arg, do arg times. (see also 'thai-backword-word)"
  (interactive "p")
  (let ((arg (or arg 1)))
    (forward-thaiword (- 0 arg))))

(defun pthai-kill-word(&optional arg)
  "kill thai word, but only match against thai words, accounting
for thai word boundaries.  kills spaces and non-thai characters
between thai words"
  (interactive "p")
  (let ((arg (or arg 1))
	(b_pt (point))
	e_pt done)
    (while (and (> arg 0) (not done))
      (cond
       ((looking-at-p "\\ct")
	(setq e_pt (cdr (pthai-bounds-of-thaiword-at-point)))
	(goto-char e_pt)
	(setq arg (1- arg)))
       ((forward-thaiword)
	(goto-char (1- (point)))
	(setq e_pt (cdr (pthai-bounds-of-thaiword-at-point)))
	(goto-char e_pt)
	(setq arg (1- arg)))
       (t
	(setq done t))))
    (while (and (< arg 0) (not done))
      (cond
       ((backward-thaiword)
	(setq e_pt (car (pthai-bounds-of-thaiword-at-point)))
	(goto-char (1- e_pt))
	(setq arg (1+ arg)))
       (t
	(setq done t))))
    (when e_pt
      (if (> e_pt b_pt)
	  (kill-region b_pt e_pt)
	(goto-char (1+ (point)))
	(kill-region e_pt b_pt)))))

(defun pthai-backward-kill-word(arg)
  "same as backward-kill-word but accommodate thai word boundaries"
  (interactive "p")
  (pthai-kill-word (- (or arg 1))))

(defun pthai-transpose-words(&optional arg)
  "similar to 'transpose-words, but accommodate thai word
boundaries, ignoring non-thai text"
  (interactive "p")
  (let ((arg (or arg 1))
	b1 b2 tmp tmp2 mark_lt)
    (cond
     ((= arg 0)
      (save-mark-and-excursion
	;; find thai words near point and mark
	(setq b1 (or (pthai-bounds-of-thaiword-at-point)
		     (and (forward-thaiword)
			  (pthai-bounds-of-thaiword-at-point (1- (point)))))
	      b2 (and (mark)
		      (goto-char (mark))
		      (or (pthai-bounds-of-thaiword-at-point)
			  (and (forward-thaiword)
			       (pthai-bounds-of-thaiword-at-point (1- (point)))))))
	(unless (and b1 b2)
	  (error "could not find 2 thai words")))
      (if (> (car b1) (car b2))
	  (setq tmp b1
		b1 b2
		b2 tmp
		mark_lt t))
      (setq tmp (buffer-substring (car b1) (cdr b1))
	    tmp2 (buffer-substring (car b2) (cdr b2)))
      (goto-char (cdr b2))
      (kill-region (car b2) (cdr b2))
      (insert tmp)
      (goto-char (cdr b1))
      (kill-region (car b1) (cdr b1))
      (insert tmp2)
      ;; move cursor to front/end of original word it was by,
      ;; depending on new position
      (goto-char (if mark_lt (car b1) (cdr b2))))
     ((> arg 0)
      (save-excursion
	;; find boundaries of thai word to move
	(setq b1 (or (pthai-bounds-of-thaiword-at-point)
		     (and (backward-thaiword)
			  (pthai-bounds-of-thaiword-at-point))))
	;; find non-thai boundaries
	(setq b2 (and b1
		      (goto-char (cdr b1))
		      (forward-thaiword)
		      (backward-thaiword)
		      (cons (cdr b1) (point))))
	;; check bounds found and move to insertion point
	(unless (and b1 b2
		     (setq tmp (and (forward-thaiword arg) (point))))
	  (error "could not find thai words")))
      (goto-char tmp)
      (insert (buffer-substring (car b2) (cdr b2)))
      (insert (buffer-substring (car b1) (cdr b1)))
      (kill-region (car b1) (cdr b2)))
     ((< arg 0)
      (save-excursion
	;; bounds of word near point
	(setq b1 (or (pthai-bounds-of-thaiword-at-point)
		     (and (backward-thaiword)
			  (pthai-bounds-of-thaiword-at-point))))
	;; non-thai bounds before b1
	(setq b2 (and b1
		      (goto-char (car b1))
		      (backward-thaiword)
		      (forward-thaiword)
		      (cons (point) (car b1))))
	(unless (and b1 b2
		     (setq tmp (and (forward-thaiword arg) (point))))
	  (error "could not find thai words")))
      (goto-char tmp)
      (setq tmp (buffer-substring (car b1) (cdr b1))
	    tmp2 (buffer-substring (car b2) (cdr b2)))
      (kill-region (car b2) (cdr b1))
      (insert tmp tmp2)
      (goto-char (- (point) (length tmp2)))))))

(defun pthai-split-region(p1 p2)
  "in a region defined by p1 and p2, split thai words into
subwords separated by spaces"
  (interactive "r")
  (let (bounds)
    (save-excursion
      (save-restriction
	(narrow-to-region p1 p2)
	(goto-char (point-min))
	;; divide up thai words with no spaces between
	(while (re-search-forward "\\(\\ct+\\)" nil t)
	  (when (and (setq bounds
			   (pthai-bounds-of-thaiwords-at-point (1- (point))))
		     (> (length bounds) 1))
	    (insert
	     (string-join
	      (mapcar (lambda (x) (buffer-substring (car x) (cdr x))) bounds)
	      " "))
	    (delete-region (caar bounds)
			   (cdar (last bounds)))))
	;; append spaces
	(goto-char (point-min))
	(while (re-search-forward "\\(\\ct\\)\\(\\Ct\\)" nil t)
	  (unless (string-match-p "[ \f\t\n\r\v]" (match-string 2))
	    (replace-match (concat (match-string 1) " " (match-string 2)))))
	;; prepend spaces
	(goto-char (point-min))
	(while (re-search-forward "\\(\\Ct\\)\\(\\ct\\)" nil t)
	  (unless (string-match-p "[ \f\t\n\r\v]" (match-string 1))
	    (replace-match (concat (match-string 1) " " (match-string 2)))))))))

(defun pthai-split(text)
  "split thai text.  return split string of text"
  (interactive (list (pthai-completing-read-dynamic "text: " nil)))
  (with-temp-buffer
    (insert text)
    (goto-char (point-min))
    (pthai-split-buffer)
    (if (called-interactively-p 'any)
	(message "%s" (buffer-string)))
    (buffer-string)))

(defun pthai-split-string-at-point()
  "split thai string at point, returns nil if point not at thai
string"
  (interactive)
  (let ((bounds (pthai-bounds-of-thaistring-at-point)))
    (and bounds
	 (pthai-split-region (car bounds) (cdr bounds)))))

(defun pthai-split-line()
  "split thai words on current line"
  (interactive)
  (pthai-split-region (line-beginning-position) (line-end-position)))

(defun pthai-split-line-from-point()
  "split thai words from point, or beginning of thai word at point,
to the end of the line"
  (interactive)
  (pthai-split-region
   (or (car (pthai-bounds-of-thaiword-at-point)) (point))
   (line-end-position)))

(defun pthai-split-buffer()
  "split thai words in buffer"
  (interactive)
  (pthai-split-region (point-min) (point-max)))

(defun pthai-unsplit-region(p1 p2)
  "remove space and tabs from thai words in region"
  (interactive "r")
  (let* ((nonnum_rexp "\\([\u0e00-\u0e4f\u0e5a-\u0e7f]\\)")
	 (space_regexp
	  (concat nonnum_rexp "\\([ \t]+\\)" nonnum_rexp)))
    (save-excursion
      (save-restriction
	(narrow-to-region p1 p2)
	(goto-char (point-min))
	;; remove spaces between thai words excluding numbers
	(while (re-search-forward space_regexp nil t)
	  (replace-match (concat (match-string 1) (match-string 3))))))))

(defun pthai-unsplit-line()
  "remove spaces from thai words on current line"
  (interactive)
  (pthai-unsplit-region (line-beginning-position) (line-end-position)))

(defun pthai-unsplit-line-from-point()
  "remove spaces from thai words on line starting at point"
  (interactive)
  (pthai-unsplit-region (point) (line-end-position)))

(defun pthai-unsplit-buffer()
  "remove spaces from thai words in buffer"
  (interactive)
  (pthai-unsplit-region (point-min) (point-max)))

(defun pthai-url-clickthai(thaiword)
  "return mp3 url for a thai word from clickthai-online.com or
nil if unavailable"
  (let ((url "https://www.clickthai-online.com/wbtde/dictionary.php")
	(url-request-method "POST")
	(url-request-extra-headers
	 `(("Content-Type" . "application/x-www-form-urlencoded; charset=utf-8")))
	(url-request-data
	 (concat "action=search&search=" (url-encode-url thaiword)))
	;; <a href="javascript:play('../cgi-bin/playmp3.pl?322476');">วัน</a>
	(click_url
	 (concat "playmp3\\.pl\\?\\([0-9]+\\)');\">" thaiword "</a>"))
	(play_url "http://www.clickthai-online.com/cgi-bin/playmp3.pl?"))
    ;; tests: ผู้ชาย วัน
    (with-current-buffer (url-retrieve-synchronously url)
      (set-buffer-multibyte t)
      (goto-char (point-min))
      (and (re-search-forward click_url nil t 1)
	   (concat play_url (match-string 1))))))

(defun pthai-url-thai-language(thaiword)
  "return mp3 url for a thai word from thai-language.com or nil
if unavailable"
  (let* ((click_regexp
	  "<a onClick=PlayAudioFile('/mp3/\\([A-Z][0-9]+\\.mp3\\)')")
	 sound_url
	 tmp
	 (mp3_url "http://thai-language.com/mp3/")
	 (url "http://thai-language.com/dict")
	 (url-request-method "POST")
	 (url-request-extra-headers
	  `(("Content-Type" . "application/x-www-form-urlencoded; charset=utf-8")))
	 (url-request-data
	  (concat "action=/default.aspx&tmode=0&search="
		  (url-encode-url thaiword))))
    ;; tests: แมว หำ วร สระ ตัว ความเสียหาย วรรค, nils บักหำ กะปิ
    (with-current-buffer (url-retrieve-synchronously url)
      (set-buffer-multibyte t)
      (goto-char (point-min))
      (while (and (not sound_url)
		  (re-search-forward click_regexp nil t))
	(and (setq tmp (cons (match-end 1) (match-string 1)))
	     (search-backward "<tr" nil t)
	     (string= thaiword
		      (replace-regexp-in-string
		       "\\Ct+" "" (buffer-substring-no-properties (point) (car tmp))))
	     (setq sound_url (concat mp3_url (cdr tmp))))
	(goto-char (car tmp)))
      sound_url)))

(defun pthai-url-thai2english(thaiword)
  "return mp3 url for a thai word from thai2english.com or nil if
unavailable"
  (let* ((t2e_url "https://www.thai2english.com")
	 (fetch_url (concat t2e_url "/search?q=" (url-encode-url thaiword)))
	 (id_match (concat t2e_url "/dictionary/\\([0-9]+\\)\\.html"))
	 (speaker_url "/img/circle-icons/speaker.svg")
	 (sounds_url "https://ast.thai2english.com/sounds/"))
    ;; tests: เป็น กรุงเทพ ประเทศไทย, nils แมว เป็นน
    (with-current-buffer (url-retrieve-synchronously fetch_url)
      (goto-char (point-min))
      (and (search-forward speaker_url nil t)
	   (goto-char (point-min))
	   (re-search-forward id_match nil t 1)
	   (concat sounds_url (match-string 1) ".mp3")))))

(defun pthai-url-lingopolo(thaiword)
  "return mp3 url for a thai word from lingopolo.org or nil if
unavailable"
  (let*
      ((lingo_url "https://lingopolo.org")
       (lingo_turl (concat lingo_url "/thai"))
       (search_url
	(concat lingo_turl "/search/node?keys=" (url-encode-url thaiword)))
       (yes_match
	(concat "\n[ \t]+<strong>" thaiword "</strong>[ \t\n\r]+\.\.\.</p>"))
       (no_match "Your search yielded no results.")
       ;; <a href="https://lingopolo.org/thai/word/also-well">also; well...</a>
       (found_regexp
	(concat "<a href=\"\\(" lingo_turl "/word/.*?\\)\">.*</a>"))
       found_url
       (audio_regexp
	"<audio src=\"\\(/thai/sites/.*?\\mp3\\)\" controls="))
    ;; tests แมว เอ้อ เสือ คน หนอนผีเสื้อ
    (with-current-buffer (url-retrieve-synchronously search_url)
      (set-buffer-multibyte t)
      (goto-char (point-min))
      (and (not (re-search-forward no_match nil t))
	   (goto-char (point-min))
	   (re-search-forward yes_match nil t)
	   (re-search-backward found_regexp nil t)
	   (setq found_url (match-string 1))))
    (when found_url
      (with-current-buffer (url-retrieve-synchronously found_url)
	(set-buffer-multibyte t)
	(goto-char (point-min))
	(and (re-search-forward audio_regexp nil t)
	     (concat lingo_url (match-string 1)))))))

(defun pthai-soundfile(thaiword)
  "generate default sound path to store mp3 audio file for a
thai word. file may or may not already exist"
  (unless (string-match-p "^\\ct+$" thaiword)
    (error "only thai words supported (%s)" thaiword))
  (unless (and pthai-audio-directory
	       (file-directory-p pthai-audio-directory))
    (error "pthai-audio-directory (%s) not found"  pthai-audio-directory))
  (let* ((subdir
	  (concat pthai-audio-directory "/"
		  (number-to-string (string-to-char thaiword))))
	 (file
	  (expand-file-name
	   (concat subdir "/" thaiword ".mp3"))))
    (unless (file-directory-p subdir)
      (make-directory subdir))
    file))

(defun pthai-soundfile-downloaded-p(thaiword)
  "return non-nil if a soundfile download already attempted for a
thai word"
  (file-exists-p (pthai-soundfile thaiword)))

(defun pthai-soundfile-playable-p(thaiword)
  "return non-nil if a soundfile downloaded successfully for a thai word"
  (and (pthai-soundfile-downloaded-p thaiword)
       (> (file-attribute-size (file-attributes (pthai-soundfile thaiword))) 0)))

(defun pthai-soundfiles-all()
  "return list of thai words for all soundfiles download attempts"
  (interactive)
  (let ((slist
	 (sort
	  (mapcar
	   (lambda(x)
	     (replace-regexp-in-string "\\.[Mm][Pp]3$" ""
				       (file-name-nondirectory x)))
	   (directory-files-recursively pthai-audio-directory "^\\ct+\\.[Mm][Pp]3$"))
	  'string-lessp)))
    (when (called-interactively-p 'any)
      (message "%s" (string-join slist " ")))
    slist))

(defun pthai-soundfiles-count()
  "return number of soundfiles downloaded"
  (interactive)
  (let ((len (length (pthai-soundfiles-all))))
    (when (called-interactively-p 'any)
      (message "%d soundfiles downloaded" len))
    len))

(defun pthai-soundfiles-unplayable()
  "return list of thai words that don't have soundfile available
after download attempt"
  (interactive)
  (let ((rlist
	 (delete nil
		  (mapcar
		   (lambda(x) (if (not (pthai-soundfile-playable-p x))
				  x))
		   (pthai-soundfiles-all)))))
    (when (called-interactively-p 'any)
      (message "%s" (string-join rlist " ")))
    rlist))

(defun pthai-soundfiles-playable()
  "return list of all thai words with downloaded soundfile"
  (interactive)
  (let ((rlist
	 (delete nil
		 (mapcar
		  (lambda(x) (if (pthai-soundfile-playable-p x)
				 x))
		  (pthai-soundfiles-all)))))
    (when (called-interactively-p 'any)
      (message "%s" (string-join rlist " ")))
    rlist))

(defun pthai-soundfiles-sort(tlist)
  "sort list of list of thai strings in order of most available
audio files as a percentage, or containing the biggest words"
  (sort tlist
	(lambda (a b)
	  (let*
	      ((a_total (length a))
	       (b_total (length b))
	       (a_found
		(length (delete nil (mapcar 'pthai-soundfile-playable-p a))))
	       (b_found
		(length (delete nil (mapcar 'pthai-soundfile-playable-p b))))
	       (a_percent (/ (float a_found) (float a_total)))
	       (b_percent (/ (float b_found) (float b_total))))
	    (cond
	     ;; use biggest length of first word if all rest same
	     ((and (= a_total b_total)
		   (= a_percent b_percent))
	      (> (length (car a)) (length (car b))))
	     ;; use biggest words if percentage same
	     ((= a_percent b_percent)
	      (< a_total b_total))
	     ;; else highest percent of words available
	     (t
	      (> a_percent b_percent)))))))

(defun pthai-mp3-play(thaiword)
  "use sox/play to play audio file for a thai word. attempts to
remove silences from audio file unless 'pthai-audio-strip-silence
is nil.  assumes soundfile already downloaded and in default
sound directory. return t if playing successful, nil or error
otherwise."
  (let ((tpath (pthai-soundfile thaiword))
	(sflags "silence 1 0.1 1% reverse silence 1 0.1 1% reverse")
	tmp)
    (cond
     ;; audio disabled
     ((not pthai-audio-play)
      nil)
     ;; check for executable
     ((not (executable-find "play"))
      (error "sox/play not found"))
     ;; check for audiofile
     ((not (pthai-soundfile-playable-p thaiword))
      (error "file not playable or existing: %s" tpath))
     ;; play audio
     (t
      (unless pthai-audio-strip-silence
	(setq sflags nil))
      (setq tmp
	    (shell-command-to-string
	     (concat "play -q \"" tpath "\" " sflags)))
      ;; xxx warnings, sound still ok?
      (if (and tmp (or (string= tmp "")
		       (string-match-p "WARN .*rate clipped" tmp)))
	  t
	(and (message tmp)
	     nil))))))

(defun pthai-mp3-url(thaiword)
  "return mp3 sound url for thai word from random one of thai-language,
clickthai, lingopolo or thai2english, or nil if not
found. returns nil unless thaiword argument given"
  (let ((fetchers
	 (pthai-randomize
	  (list (cons pthai-fetch-clickthai 'pthai-url-clickthai)
		(cons pthai-fetch-lingopolo 'pthai-url-lingopolo)
		(cons pthai-fetch-thai2english 'pthai-url-thai2english)
		(cons pthai-fetch-thailanguage 'pthai-url-thai-language))))
	 url)
    (when (string-match-p "\\ct+" thaiword)
      (catch 'fdone
	(dolist (tmp fetchers)
	  (if (and (car tmp)
		   (setq url (funcall (cdr tmp) thaiword)))
	      (throw 'fdone url)))))))

(defun pthai-mp3-fetch(thaiword &optional force_dl)
  "fetch mp3 soundfile to default audio directory if file missing
or forced to. returns path to downloaded file or nil if unable to
download. if download unsuccesful, write empty file with name of
thai word to default path.  if downloading disabled and non-empty
file already downloaded, return full path"
  (let ((file (pthai-soundfile thaiword))
	url)
    (unless (string-match-p "\\ct+" thaiword)
      (error "thaiword %s contains no thai chars"))
    (cond
     ;; file already downloaded but downloading disabled
     ((and (not pthai-audio-download)
	   (pthai-soundfile-playable-p thaiword))
      file)
     ;; downloading enabled,
     ;; and download non-dict words enabled or word in dict,
     ;; forced download or file not downloaded,
     ;; try to download and return path if successful
     ((and pthai-audio-download
	   (or pthai-audio-download-non-dictionary (pthai-dictionary-p thaiword))
	   (or force_dl (not (pthai-soundfile-downloaded-p thaiword))))
      (if (and (setq url (pthai-mp3-url thaiword))
	       (url-copy-file url file t))
	  file
	(and (write-region "" nil file nil 1)
	     nil)))
     ;; otherwise, no downloading and return whether file available
     (t
      (if (pthai-soundfile-playable-p thaiword)
	  file nil)))))

(defun pthai-download-and-play(word &optional force_dl)
  "download and play audio for a thai word. returns t if played
complete audio for word, nil otherwise"
  (unless (string-match-p "^\\ct+$" word)
    (error "only thai words supported"))
  ;; most common cases: existing word there and available for
  ;; download/download successful
  (if (and (pthai-mp3-fetch word force_dl) (pthai-mp3-play word))
      t
    ;; split up word download and play "best' word combo
    (let* ((splits (pthai-split-all word))
	   (tmp (pthai-soundfiles-sort (mapcar 'split-string splits)))
	   dl_done play_done)
      ;; try to download set that makes up the word
      (when pthai-audio-download
	(dolist (tmp (pthai-soundfiles-sort (mapcar 'split-string splits)))
	  (and (not dl_done)
	       (mapc (lambda(x) (pthai-mp3-fetch x force_dl)) tmp)
	       (seq-every-p 'pthai-soundfile-playable-p tmp)
	       (setq dl_done tmp))))
      ;; find combo with most downloads and play
      (when pthai-audio-play
	(setq play_done (or dl_done
			    (car (pthai-soundfiles-sort
				  (mapcar 'split-string splits)))))
	(setq play_done
	      (mapcar (lambda (x)
			(and (pthai-soundfile-playable-p x)
			     ;; skip single consonants
			     (not (string-match-p pthai-regexp-single-consonant x))
			     (pthai-mp3-play x)))
		      play_done)))
      (if (member nil play_done)
	    nil t))))

(defun pthai-highlight-say-bounds(bounds &optional force_dl no_audio)
  "helper to highlight and say word at boundary, lookup and
message its definition. if option force_dl, force audio download.
if option no_audio, don't play audio"
  (let* ((p1 (car bounds))
	 (p2 (cdr bounds))
	 (word (buffer-substring-no-properties p1 p2)))
    (when (and p1 p2 word)
      ;; xxx double display of definitions because when download
      ;; required display of definition is cleared
      (cond
       ;; number
       ((string-match-p pthai-regexp-any-number word)
	(pulse-momentary-highlight-region p1 p2)
	(pthai-lookup word t)
	(redisplay)
	(or no_audio (pthai-say-number word))
	(pthai-lookup word t))
       ;; thai word
       ((string-match-p "^\\ct+$" word)
	(pulse-momentary-highlight-region p1 p2)
	(pthai-lookup word t)
	(redisplay)
	(or no_audio (pthai-download-and-play word force_dl))
	(pthai-lookup word t))
       ;; english word
       ((string-match-p "^\\([a-zA-Z]+\\)$" word)
	(pulse-momentary-highlight-region p1 p2)
	(redisplay)
	(if no_audio
	    (pthai-lookup word t)
	  (dolist (tmp (pthai-lookup word))
	    (pthai-download-and-play tmp force_dl)
	    (pthai-lookup word t))))))))

(defun pthai-say-region-helper(p_begin p_end &optional force_dl thai_only)
  "helper function to play audio for a selected region. force
downloads with force_dl arg. play only thai words with thai_only
arg"
  (let ((old_cursor cursor-type)
	(m_rexp (concat "\\("
			"\\ct+\\|"
			"[0-9]+"
			(if thai_only "\\)"
			  "\\|[a-zA-Z]+\\)")))
	bounds)
    (unwind-protect
	(save-mark-and-excursion
	  (goto-char p_begin)
	  (deactivate-mark)
	  (setq cursor-type nil)
	  (while (re-search-forward m_rexp p_end t)
	    (setq bounds
		  (or (pthai-bounds-of-thaiwords-at-point (match-beginning 1))
		      (list (cons (match-beginning 1) (match-end 1)))))
	    ;; filter words that are completely out of the region, which can
	    ;; happen if thai string overlaps end points
	    (setq bounds
		  (seq-remove (lambda (x) (or (>= p_begin (cdr x))
					      (>= (car x) p_end)))
			      bounds))
	    (dolist (tmp bounds)
	      (pthai-highlight-say-bounds tmp force_dl))))
      (setq cursor-type old_cursor))))

(defun pthai-say(text &optional force_dl thai_only)
  "split up text string, download and play audio for thai
words. return definitions of words as a string.  also say arabic
or thai numbers in thai.  see 'pthai-split for more info
on how words are split"
  (interactive (list (pthai-completing-read-dynamic "word: " nil)))
  (let (words)
    (cond
     ;; meanings of english words with spaces, e.g., "good morning"
     ((and (string-match-p "^[a-zA-Z ]+$" text)
	   (setq words (or (pthai-lookup text)
			   (pthai-lookup-classifier text))))
      (message "%s = %s" text (string-join words ", "))
      (dolist (tmp words)
	(pthai-download-and-play tmp force_dl)))
     (t
      (save-excursion
	(with-temp-buffer
	  (insert text)
	  (pthai-say-region-helper (point-min)
				   (point-max) force_dl thai_only)))))))

(defun pthai-say-region(p1 p2 &optional force_dl)
  "play thai audio for selected region, force downloads if option force_dl"
  (interactive "r\nP")
  (pthai-say-region-helper p1 p2 force_dl))

(defun pthai-say-region-thai-only(p1 p2 &optional force_dl)
  "play thai audio for thai words in highlighted region"
  (interactive "r\nP")
  (pthai-say-region-helper p1 p2 force_dl t))

(defun pthai-say-buffer(&optional force_dl)
  "play thai audio for buffer"
  (interactive "P")
  (pthai-say-region-helper (point-min) (point-max) force_dl))

(defun pthai-say-buffer-from-point(&optional force_dl)
  "play thai audio for buffer from current point"
  (interactive "P")
  (pthai-say-region-helper (point) (point-max) force_dl))

(defun pthai-say-buffer-from-point-thai-only(&optional force_dl)
  "play thai audio for buffer from current point, thai words only"
  (interactive "P")
  (pthai-say-region-helper (point) (point-max) force_dl t))

(defun pthai-say-buffer-thai-only(&optional force_dl)
  "play thai audio for thai words in buffer"
  (interactive "P")
  (pthai-say-region-helper (point-min) (point-max) force_dl t))

(defun pthai-say-line(&optional force_dl)
  "play thai audio for current line in buffer"
  (interactive "P")
  (pthai-say-region-helper (line-beginning-position)
			   (line-end-position)
			   force_dl))

(defun pthai-say-line-thai-only(&optional force_dl)
  "play thai audio for current line in buffer, thai words only"
  (interactive "P")
  (pthai-say-region-helper (line-beginning-position)
			   (line-end-position)
			   force_dl
			   t))

(defun pthai-say-string-at-point(&optional force_dl)
  "play thai audio for thai string (a thai text string separated
by whitespace) at point in buffer"
  (interactive "P")
  (let* ((tmp (pthai-bounds-of-thaistring-at-point))
	 (p1 (car tmp))
	 (p2 (cdr tmp)))
    (unless (and p1 p2)
      (error "could not determine bounds"))
    (pthai-say-region-helper p1 p2 force_dl)))

(defun pthai-say-word-at-point(&optional force_dl)
  "play thai audio for thing at point.  error if bounds cannot be
determined"
  (interactive "P")
  (let ((bounds
	 (cond
	  ;; thai word
	  ((looking-at-p "\\ct")
	   (pthai-bounds-of-thaiword-at-point))
	  ;; on space at end of thai word
	  ((and (or (looking-at-p "[ \f\t\n\r\v]") (eobp))
		(string-match-p "\\ct" (char-to-string (preceding-char))))
	   (pthai-bounds-of-thaiword-at-point (1- (point))))
	  ;; eng word
	  ((looking-at-p "[A-Za-z]")
	   (bounds-of-thing-at-point 'word))
	  ;; number
	  ((looking-at-p "[0-9๑๒๓๔๕๖๗๘๙๐]")
	   (bounds-of-thing-at-point 'word))
	  (t
	   nil))))
    (unless bounds
      (error "could not determine bounds"))
    (pthai-highlight-say-bounds bounds force_dl)))

(defun pthai-say-number(num)
 "say number (thai or arabic) in thai. if arabic number, first
convert to thai. returns thai number (string), or nil if can't
convert.  only integers < abs(10 million) are supported.  use
'pthai-say for interactive usage"
  (let* ((powers '(("1" . "เอ็ด") ("10" . "สิบ") ("20" . "ยี่สิบ")
		  ("100" . "ร้อย") ("1000" . "พัน") ("10000" . "หมื่น")
		  ("100000" . "แสน") ("1000000" . "ล้าน")))
	(fn2w (lambda (x) (cdr (assoc x powers))))
	(negative "ลบ")
	(max_num 9999999)
	is_neg
	anum
	rnum
	tlist
	tmp
	power
	n)
    (unless (stringp num)
      (setq num (number-to-string num)))
    ;; negative number
    (when (string-match-p "^-" num)
      (setq num (replace-regexp-in-string "^-" "" num))
      (setq is_neg t))
    (unless (string-match-p pthai-regexp-any-number num)
      (error (concat "not valid number " num)))
    (if (string-match-p pthai-regexp-thai-number num)
	(setq num (pthai-number num)))
    (unless (string-match-p "^[0-9]+$" num)
      (setq num (replace-regexp-in-string "[^0-9]" "" num)))
    (setq anum (string-to-number num))
    (cond ((> anum max_num)
	   (warn (concat "audio only for numbers <= abs("
			 (number-to-string max_num) ")")))
	  ((< anum 10)
	   (setq tlist (list anum)))
	  (t
	   (setq rnum (reverse (split-string (pthai-number num) "" t)))
	   ;; digit 1
	   (setq tmp (nth 0 rnum))
	   (cond ((string= "๑" tmp)
		  (setq tlist (cons (funcall fn2w "1") tlist)))
		 ((not (string= "๐" tmp))
		  (setq tlist (cons tmp tlist))))
	   ;; digit 2
	   (when (setq tmp (nth 1 rnum))
	     (cond ((string= "๑" tmp)
		    (setq tlist (cons (funcall fn2w "10") tlist)))
		   ((string= "๒" tmp)
		    (setq tlist (cons (funcall fn2w "20") tlist)))
		   ((not (string= "๐" tmp))
		    (setq tlist (cons (funcall fn2w "10") tlist))
		    (setq tlist (cons tmp tlist)))))
	   ;; digits 3-7
	   (setq power "10")
	   (dotimes (n 5)
	     (when (setq tmp (nth (+ n 2) rnum))
	       (setq power (concat power "0"))
	       (when (not (string= tmp "๐"))
		 (setq tlist (cons (funcall fn2w power) tlist))
		 (setq tlist (cons tmp tlist)))))))
    ;; convert all to strings
    (setq tlist
	  (mapcar (lambda (x) (if (numberp x) (number-to-string x) x))
		  tlist))
    ;; convert all arabic numbers to thai
    (setq tlist
	  (mapcar (lambda (x) (if (string-match-p "^[0-9]+" x)
				  (pthai-number x) x))
		  tlist))
    ;; negative
    (if is_neg (setq tlist (cons negative tlist)))
    ;; say the number
    (dolist (tmp tlist)
      (pthai-download-and-play tmp))
    ;; return thai number as a string
    (setq tmp (pthai-number num))
    (if is_neg (setq tmp (concat "-" tmp)))
    (if (called-interactively-p 'any)
	(message tmp))
    tmp))

(defun pthai-number(&optional num)
  "convert a number between thai and arabic, and vice versa.
message user converted value if called interactively"
  (interactive)
  (let ((nlist '(("0" . "๐") ("1" . "๑") ("2" . "๒") ("3" . "๓")
		 ("4" . "๔") ("5" . "๕") ("6" . "๖") ("7" . "๗")
		 ("8" . "๘") ("9" . "๙")))
	numa
	res)
    (unless num
      (setq num (read-string "number: ")))
    (unless (stringp num)
      (setq num (number-to-string num)))
    (setq numa (split-string num "" t))
    (if (string-match-p "[0-9]+" num)
	(setq res
	      (mapconcat (lambda(x) (or (cdr (assoc x nlist)) x)) numa ""))
      (setq res
	    (mapconcat (lambda(x) (or (car (rassoc x nlist)) x)) numa "")))
    (if (called-interactively-p 'any)
	(message res))
    res))

(defun pthai-parse-hour-minute-second(timestr &optional thaitime)
  "parse hour-minute-second strings like 7 1pm 12:32pm 22:07
9:45p.m. 0645 3.27am 15:23:55 or \"now\".  returns list of (SEC
MIN HOUR), with nil for each element not parsed"
  (let ((ampm_rexp "\\([AaPp]\.?[Mm]\.?\\)$")
	am_pm second minute hour tlist)
    ;; "now"
    (if (string-match-p "^[Nn][Oo][Ww]$" timestr)
	(setq timestr
	      (nth 3 (split-string
		      (pthai-time-now (not thaitime))))))
    ;; strip out anything that doesn't look like a time
    (setq timestr
	  (replace-regexp-in-string "[^0-9:\.pPaAmM]" "" timestr))
    ;; determine am/pm and remove from string
    (when (string-match-p ampm_rexp timestr)
      (if (string-match-p "[Aa]" timestr)
	  (setq am_pm "am")
	(setq am_pm "pm"))
      (setq timestr
	    (replace-regexp-in-string ampm_rexp "" timestr)))
    ;; parse hour/minute/seconds digits
    (cond
     ;; 1 or 2 digit, 1 11
     ((string-match "^\\([0-9]\\{1,2\\}\\)$" timestr)
      (setq hour (match-string 1 timestr))
      (setq minute "0")
      (setq second "0"))
     ;; 4 digits, 0645
     ((string-match "^\\([0-9]\\{4\\}\\)$" timestr)
      (setq hour (substring timestr 0 2))
      (setq minute (substring timestr 2 4))
      (setq second "0"))
     ;; 6 digits, 064532
     ((string-match "^\\([0-9]\\{6\\}\\)$" timestr)
      (setq hour (substring timestr 0 2))
      (setq minute (substring timestr 2 4))
      (setq second (substring timestr 4 6)))
     ;; 12:32 22:07 3.27 2:35
     ((string-match "^\\([0-9]\\{1,2\\}\\)[:\.]\\([0-9]\\{1,2\\}\\)$" timestr)
      (setq hour (match-string 1 timestr))
      (setq minute (match-string 2 timestr))
      (setq second "0"))
     ;; 12:43:54 or 12.43.54
     ((string-match
       "^\\([0-9]\\{1,2\\}\\)[:\.]\\([0-9]\\{1,2\\}\\)[:\.]\\([0-9]\\{1,2\\}\\)$" timestr)
      (setq hour (match-string 1 timestr))
      (setq minute (match-string 2 timestr))
      (setq second (match-string 3 timestr))))
    ;; convert to strings
    (if second (setq second (string-to-number second)))
    (if minute (setq minute (string-to-number minute)))
    (if hour (setq hour (string-to-number hour)))
    ;; adjust/sanity checks for am/pm
    (when (and am_pm hour)
      (cond
       ((and am_pm (> hour 12))
	(setq hour nil))
       ((and (string= am_pm "am") (= 12 hour))
	(setq hour 0))
       ((and (string= am_pm "pm") (< hour 12))
	(setq hour (+ 12 hour)))))
    ;; return nil unless all seem valid
    (if (and hour (>= hour 0) (<= hour 24)
	     minute (>= minute 0) (< minute 60)
	     second (>= second 0) (< second 60))
	(list second minute hour)
      (list nil nil nil))))

(defun pthai-time-loc2th(timestr &optional reverse)
"convert an hour:minute:second time string to thai time. use
\"now\" for current time.  if optional reverse, convert thai time
to local time"
  (interactive "stime: ")
  (let* ((c_time (current-time))
	 (time_format (concat "%a %b %e %H:%M:%S %Z %Y"))
	 (thai_offset (* 7 3600)) ;; UTC+7 or ICT or +07 or "Asia/Bangkok"
	 (thai_zone (current-time-zone nil thai_offset))
	 tlist enc_time loc_fts thai_fts o_str)
    ;; keep reading until valid second/minute/hour
    (while (not
	    (and (setq tlist (pthai-parse-hour-minute-second timestr reverse))
		 (nth 0 tlist)
		 (nth 1 tlist)
		 (nth 2 tlist)))
      (setq timestr (read-string "time: ")))
    (cond
     ;; thai time
     (reverse
      (setq tlist (append tlist (nthcdr 3 (decode-time c_time thai_zone))))
      (setq enc_time (apply 'encode-time tlist))
      (setq loc_fts (format-time-string time_format enc_time (current-time-zone)))
      (setq thai_fts (format-time-string time_format enc_time thai_zone))
      (setq o_str (concat "thai time = " thai_fts "\n" "time      = " loc_fts)))
     ;; local time
     (t
      (setq tlist (append tlist (nthcdr 3 (decode-time c_time))))
      (setq enc_time (apply 'encode-time tlist))
      (setq loc_fts (format-time-string time_format enc_time (current-time-zone)))
      (setq thai_fts (format-time-string time_format enc_time thai_zone))
      (setq o_str (concat "time      = " loc_fts "\n" "thai time = " thai_fts))))
    (message o_str)))

(defun pthai-time-th2loc(timestr)
"convert a thai time (hour/minute/second) to local time.  use
\"now\" for current time"
  (interactive "stime: ")
  (pthai-time-loc2th timestr t))

(defun pthai-time-now(&optional loctime)
  "return thai current time.  if optional loctime, return current
local time. message user time if called interactively"
  (interactive "P")
  (let ((c_zone (current-time-zone))
	(t_zone (current-time-zone nil (* 7 3600)))
	(time_format (concat "%a %b %e %H:%M:%S %Z %Y"))
	tstr)
    (setq tstr
	  (format-time-string time_format (current-time)
			      (if loctime c_zone t_zone)))
    (if (called-interactively-p 'any)
	(message tstr))
    tstr))

(defun pthai-say-time(timestr)
  "say a time in thai, use time like 1pm, 12:45pm, 17:34.  use
\"now\" for current time. returns time string"
  (interactive "stime: ")
  (let (tlist hour minute mhour tmp)
    (while (not (and (setq tlist (pthai-parse-hour-minute-second timestr))
		     (setq minute (nth 1 tlist))
		     (setq hour (nth 2 tlist))))
      (setq timestr (read-string "time: ")))
    (setq mhour (mod hour 6))
    ;; get hour
    (cond
     ;; 12am = เที่ยงคืน or หกทุ่ม or สองยาม
     ((or (< hour 1) (and (= hour 24) (= minute 0)))
      (setq tmp (random 4))
      (cond ((< tmp 2)
	     (setq tlist (list "เที่ยงคืน")))
	    ((= tmp 2)
	     (setq tlist (list "หกทุ่ม")))
	    (t
	     (setq tlist (list "สองยาม")))))
     ;; 1am 2am 3am 4am 5am = ตี + number[1-5]
     ((< hour 6)
      (setq tlist (list "ตี" hour)))
     ;; 6am = หก โมง เช้า or ตีหก or ย่ำรุ่ง
     ((< hour 7)
      (setq tmp (random 4))
      (cond ((< tmp 2)
	     (setq tlist (list "หก" "โมง" "เช้า")))
	    ((= tmp 2)
	     (setq tlist (list "ตีหก")))
	    (t
	     (setq tlist (list "ย่ำรุ่ง")))))
     ;; 7am 8am 9am 10am 11am = 1-5  + โมง + เช้า
     ;;                       = 7-11 + โมง + (เช้า)
     ((< hour 12)
      (if (= 0 (random 2))     ;; 1-5
	  (if (= mhour 1)
	      (setq tlist (list "โมง" "เช้า"))
	    (setq tlist (list mhour "โมง" "เช้า")))
	(setq tlist (list hour "โมง"))  ;; 7-11
	(if (= 0 (random 2))
	    (setq tlist (append tlist (list "เช้า"))))))
     ;; 12pm = เที่ยงวัน
     ((< hour 13)
      (setq tlist (list "เที่ยงวัน")))
     ;; 1pm 2pm 3pm 4pm 5pm   = บ่าย + number[1-5] + โมง or
     ;;             4pm       = สี่ +  โมง + เย็น or
     ;;             5pm       = ห้า + โมง + เย็น
     ((< hour 18)
      (cond
       ((= hour 13) ;; 1pm
	(setq tlist (list "บ่าย" "โมง")))
       ((or (= hour 14) (= hour 15))
	(setq tlist (list "บ่าย" mhour "โมง")))
       ((= hour 16) ;; 4pm
	(if (= 0 (random 2))
	    (setq tlist (list "สี่" "โมง" "เย็น"))
	  (setq tlist (list "บ่าย" mhour "โมง"))))
       (t           ;; 5pm
	(if (= 0 (random 2))
	    (setq tlist (list "ห้า" "โมง" "เย็น"))
	  (setq tlist (list "บ่าย" mhour "โมง"))))))
     ;; 6pm = หก โมง เย็น or ย่ำค่ำ
     ((< hour 19)
      (if (= 0 (random 3))
	  (setq tlist (list "ย่ำค่ำ"))
	(setq tlist (list "หก" "โมง" "เย็น"))))
     ;; 7pm 8pm 9pm 10pm 11pm = number[1-5] + ทุ่ม
     ((< hour 24)
      (setq tlist (list mhour "ทุ่ม")))
     (t
      (error (concat "bad hour " hour))))
    ;; minutes are appended 10:07am = สิบ โมง (เช้า) + เจ็ด นาที
    (unless (or (= minute 0) (= hour 24))
      (setq tlist (append tlist (list minute "นาที"))))
    ;; convert everything to strings
    (setq tlist (mapcar (lambda (x) (if (numberp x) (number-to-string x) x))
			tlist))
    ;; play audio
    (dolist (tmp tlist)
      (pthai-say tmp))
    (setq tmp
	  (concat (number-to-string hour) ":" (format "%02d" minute)))
    (message tmp)
    tmp))

(defun pthai-spell-p(thaistr)
  "return non-nil if thai string (a word or contiguous thai
words) is spelled correctly, i.e., string or strings are in dictionary"
  (if (pthai-dictionary-p thaistr)
      t
    (let* ((slen (length thaistr))
	   (dp (make-bool-vector (1+ slen) nil)))
      (unless (> slen 0)
	(error "no word to check"))
      (aset dp 0 t)
      (dolist (i (number-sequence 1 slen))
	(catch 'psc-done
	  (dotimes (j i)
	    (when (and (aref dp j)
		       (pthai-dictionary-p (substring thaistr j i)))
	      (aset dp i t)
	      (throw 'psc-done t)))))
      (aref dp slen))))

(defun pthai-spell-reducer(slist)
  "helper function to reduce spelling suggestions, 
converts ((ไม่ มี) (t t)) => ((ไม่มี) (t))"
  (if (or (< (length (nth 1 slist)) 2)
	  (< (length (seq-filter (lambda(x) (eq x t)) (nth 1 slist))) 2))
      slist
    (let ((n_spl (nth 0 slist))
	  (n_cho (nth 1 slist))
	  (i 0) tmp)
      (while (< i (1- (length n_cho)))
	(if (and (eq t (nth i n_cho))
		 (eq t (nth (1+ i) n_cho)))
	    (setq tmp (concat (nth i n_spl) (nth (1+ i) n_spl))
		  n_spl (pthai-nth-set tmp i n_spl)
		  n_spl (pthai-nth-delete (1+ i) n_spl)
		  n_cho (pthai-nth-delete i n_cho))
	  (setq i (1+ i))))
      (list n_spl n_cho))))

(defun pthai-spell-suggest(word)
  "return a list of suggested thai spellings for a thai word
based on levenshtein distance.  return nil if word is contained
in dictionary or no suggestions are available"
  (let ((lev_dist '(1 2 3 4))
	lev done rval)
    (cond
     ;; misspelling and no string-distance
     ((and (< emacs-major-version 27) (not (pthai-dictionary-p word)))
      (warn "suggestions unavailable, 'string-distance unimplemented for emacs versions < 27 ")
      nil)
     ;; spelled correctly, no suggestions
     ((pthai-dictionary-p word)
      nil)
     (t
      (while (and (not done)
		  (setq lev (car lev_dist)))
	(setq lev_dist (cdr lev_dist))
	(dolist (elem (hash-table-keys pthai-dictionary))
	  (if (and (<= (string-distance word elem) lev)
		   (not (string-match-p pthai-regexp-any-number elem)))
	      (push elem rval)))
	(if (> (length rval) 0)
	    (setq done t)))
      ;; xxx sort by word frequencies for better matches
      (sort rval 'string-lessp)))))

(defun pthai-spell-merge-singles(slist)
  "reduce single char splits trying to get better spell
suggestions, for example convert 'เพื่อ อ ไร' => 'เพื่อ อไร'"
  (if (< (length slist) 2)
      slist
    (let ((rlist slist)
	  (i 0) a b tmp)
      (while (< i (1- (length rlist)))
	(if (or (pthai-dictionary-p (nth i rlist))
		(> (length (nth i rlist)) 1)
		(string-match-p pthai-regexp-valid-standalone-char (nth i rlist)))
	    (setq i (1+ i))
	  (cond
	   ((= i 0)
	    (setq a i b (1+ i)))
	   ((= i (1- (length rlist)))
	    (setq a (1- i) b i))
	   ((not (pthai-dictionary-p (nth (1- i) rlist)))
	    (setq a (1- i) b i))
	   (t ;;(not (pthai-dictionary-p (nth i rlist)))
	    (setq a i b (1+ i))))
	  (setq tmp (concat (nth a rlist) (nth b rlist))
		rlist (pthai-nth-set tmp a rlist)
		rlist (pthai-nth-delete b rlist))))
      rlist)))

(defun pthai-spell-suggestions(tstr &optional no_split no_reduce keep_nil keep_singles)
  "split thai word and return list of word splits/and spelling
suggestions from dictionary where subwords are t if in dict, list
of word spelling suggestions, or nil of no/too many suggestions, e.g.,
 เตีอง => (((เ ตี อง) (nil t t))
         ((เตี อง)  ((ตี สตี เต เตก เตช เตย เตะ เตา เลตี) t))
         ((เตีอง)   ((เตียง))))
option no_split implies no word splitting done and suggestions
are for a single thai word.  suggestions with nil choices are
filtered out, unless optional keep_nil or filtering out would
leave no suggestions.  option no_reduce disables consolidation of
suggestions, see 'pthai-spell-reducer. option keep_singles to
disable merging of splits of single characters"
  (let (slist sug_list flist)
    ;; split string
    (setq slist (if no_split (list (list tstr))
		  (mapcar 'split-string (pthai-split-all tstr))))
    ;; merge invalid stand alone characters to try for better word
    ;; suggestions
    (unless keep_singles
      (setq slist (seq-uniq (mapcar 'pthai-spell-merge-singles slist))))
    ;; get spell suggestions
    (setq sug_list (mapcar
		    (lambda(x)
		      (mapcar
		       (lambda(y) (or (pthai-spell-p y)
				      (pthai-spell-suggest y)))
		       x))
		    slist))
    ;; map splitting to selections
    (setq slist (seq-map-indexed
		 (lambda(elt idx) (list elt (nth idx sug_list)))
		 slist))
    ;; eliminate "equivalent" combinations
    (unless no_reduce
      (setq slist (seq-uniq (mapcar 'pthai-spell-reducer slist))))
    ;; filter out splittings with invalid selections if other choices are available
    (and (not keep_nil)
	 (setq flist (seq-filter
		      (lambda(x) (if (memq nil (nth 1 x)) nil t))
		      slist))
	 (> (length flist) 0)
	 (setq slist flist))
    slist))

;; xxx sample misspellings, เพราะพรุ่งนี้ได้หยุดเสาร์ทิตย์ บ้านพี่สกลหนาวเเล้ว เตีอง เพื่ออไร
(defun pthai-spell-sort(slist)
  "sort array of split thai text strings by misspellings and
'best' suggestions. car of list should be \"best\" suggestion.
slist is a list of
	((split-words) (list-of-suggestions/split-word)))."
  (sort slist
	(lambda (a b)
	  (let* ((a_sug (nth 1 a))
		 (b_sug (nth 1 b))
		 (fu_all_t
		  (lambda(x)
		    (seq-every-p (lambda(y) (eq y t)) x)))
		 (a_t (funcall fu_all_t a_sug))
		 (b_t (funcall fu_all_t b_sug))
		 (a_all_sug (seq-every-p 'listp a_sug))
		 (b_all_sug (seq-every-p 'listp b_sug))
		 (a_has_nil (memq nil a_sug))
		 (b_has_nil (memq nil b_sug))
		 (a_t_or_sug (not a_has_nil))
		 (b_t_or_sug (not b_has_nil))
		 (fu_words_per_sug
		  (lambda(x) (/ (float (apply '+ (mapcar 'length (seq-filter 'listp x))))
				(max 1.0 (float (length (seq-filter 'listp x)))))))
		 (a_words_per_sug (funcall fu_words_per_sug a_sug))
		 (b_words_per_sug (funcall fu_words_per_sug b_sug))
		 (fu_avelen_per_sug_word
		  (lambda(x)
		    (/ (float (apply '+ (mapcar (lambda(y) (apply '+ y))
						(mapcar (lambda (z) (mapcar 'length z))
							(seq-filter 'listp x)))))
		       (max 1.0 (float (apply '+(mapcar 'length (seq-filter 'listp x))))))))
		 (a_avelen_per_sug_word (funcall fu_avelen_per_sug_word a_sug))
		 (b_avelen_per_sug_word (funcall fu_avelen_per_sug_word b_sug))
		 )
	    (cond
	     ;; a found for all, b not
	     ((and a_t (not b_t))
	      t)
	     ;; b found for all, a not
	     ((and (not a_t) b_t)
	      nil)
	     ;; a has no nils, b has
	     ((and (not a_has_nil) b_has_nil)
	      t)
	     ;; b has no nils, a has
	     ((and (not b_has_nil) a_has_nil)
	      t)
	     ;; a 1 word with a few sug
	     ((and a_all_sug
		   (> 3 (length (car a_sug)))
		   (= 1 (length (nth 0 a))))
	      t)
	     ;; b 1 word with a few sug
	     ((and b_all_sug
		   (> 3 (length (car b_sug)))
		   (= 1 (length (nth 0 b))))
	      nil)
	     ;; a has sug & t, b just sug
	     ((and a_t_or_sug b_all_sug)
	      t)
	     ;; b has sug & t, a only sug
	     ((and b_t_or_sug a_all_sug)
	      nil)
	     ;; length a < b
	     ((< (length a_sug) (length b_sug))
	      t)
	     ;; length b < a
	     ((< (length b_sug) (length a_sug))
	      nil)
	     ;; a ave size per word suggestion > b / vice versa
	     ((and (> a_avelen_per_sug_word 0.0)
		   (> b_avelen_per_sug_word 0.0))
	      (> a_avelen_per_sug_word b_avelen_per_sug_word))
	     ;; a has fewer words per suggestions / vice versa
	     ((and (> a_words_per_sug 0.0)
		   (> b_words_per_sug 0.0))
	      (< a_words_per_sug b_words_per_sug))
	     ;; default
	     (t
	      t))))))

(defun pthai-spell-region(p1 p2 &optional no_split)
  "spell thai words in region p1 p2, if option no_split, assume
spelling a single word with no word splitting done"
  (interactive "r\nP")
  (let* ((case-fold-search nil)
	 (n_pt (point))
	 (p_reporter
	  (make-progress-reporter "Checking..." p1 p2 nil 5 0.5))
	 (s_choice
	  (list '(?\ "skip" "skip word")
		'(?r "replace word" "replace word")
		'(?R "Replace str" "replace string")
		'(?a "accept word" "accept word for session")
		'(?A "Accept str" "accept string for session")
		'(?d "dict regen" "regenerate dictionary")
		'(?x "exit" "exit and move cursor to original point")
		'(?X "eXit at point" "exit and leave cursor at current point")))
	 (split_key ?c)
	 (key_choice
	  (seq-difference
	   (mapcar 'string-to-char
		   (append (mapcar 'number-to-string (number-sequence 0 9))
			   (mapcar 'char-to-string (number-sequence ?a ?z))
			   (mapcar 'char-to-string (number-sequence ?A ?Z))))
	   (cons split_key (mapcar 'car s_choice))))
	 (hl_on (lambda(b e)
		  (overlay-put (make-overlay b e) 'face '(:background "#FFFFAA"))))
	 (hl_off (lambda(b e)
		   (remove-overlays b e)))
	 s_bound s_str
	 s_all s_idx s_idx_change
	 (f_splits (lambda() (car (nth s_idx s_all))))
	 (f_choices (lambda(&optional x) (cadr (nth (or x s_idx) s_all))))
	 w_bound w_choice t_bound
	 ch_select l_select
	 c_done c_quit tmp)
    (goto-char p1)
    (while (and (< (point) p2)
		(not c_quit)
		(re-search-forward pthai-regexp-spelling-chars p2 t)
		(setq s_bound (cons (match-beginning 1) (match-end 1))
		      s_str (buffer-substring-no-properties (car s_bound)
							    (cdr s_bound))))
      (goto-char (cdr s_bound))
      (progress-reporter-update p_reporter (point))
      (unless (pthai-spell-p s_str)
	(setq c_done nil
	      s_idx 0
	      s_all (pthai-spell-sort (pthai-spell-suggestions s_str no_split)))
	;; double check a solution was not really found with different
	;; word-splitting, should be first choice
	(unless (seq-every-p (lambda(x) (eq x t)) (funcall f_choices 0))
	  (while (not c_done)
	    ;; update indices for choices if needed
	    (if s_idx_change
		(setq s_idx_change nil
		      s_idx (% (1+ s_idx) (length s_all))))
	    ;; find bounds of first misspelled word
	    (if (not s_all)
		(setq w_bound s_bound w_choice nil)
	      (setq w_choice nil w_bound nil)
	      (dotimes (i (length (funcall f_splits)))
		(when (and (not w_bound)
			   (not (eq t (nth i (funcall f_choices)))))
		  (setq w_choice (nth i (funcall f_choices)))
		  (setq tmp (pthai-bounds-offset-of-split-string (car s_bound) (funcall f_splits)))
		  (setq w_bound (nth i tmp)))))
	    (unless w_bound
	      (error "could not determine bounds for region"))
	    ;; map word selections to available key choices, put definitions in help menu
	    (setq tmp nil)
	    (dotimes (i (min (length w_choice) (length key_choice)))
	      (setq tmp (cons (list (nth i key_choice)
				    (nth i w_choice)
				    (string-join (pthai-lookup (nth i w_choice)) ", "))
			      tmp)))
	    (setq w_choice (reverse tmp))
	    ;; add menu option to resplit if more than 1 splits
	    (setq tmp (and (> (length s_all) 1)
			   (list
			    (list split_key
				  (concat "cycle-splits(" (number-to-string (1+ s_idx)) "/"
					  (number-to-string (length s_all)) ")")
				  "view other splits"))))
	    ;; goto end of misspelled word, highlight word, and prompt user
	    (goto-char (cdr w_bound))
	    (funcall hl_on (car w_bound) (cdr w_bound))
	    (unwind-protect
		(setq l_select (read-multiple-choice "" (append w_choice s_choice tmp)))
	      (funcall hl_off (car w_bound) (cdr w_bound)))
	    (setq ch_select (car l_select))
	    (setq c_done
		  (cond
		   ;; leave word as is and goto next
		   ((char-equal ?\  ch_select)
		    t)
		   ;; menu choice to replace word
		   ((memq ch_select (mapcar 'car w_choice))
		    (delete-region (car w_bound) (cdr w_bound))
		    (setq tmp (cadr l_select))
		    (insert tmp)
		    (goto-char (- (point) (length tmp)))
		    (setq p2 (+ p2 (- (- (cdr w_bound) (car w_bound))) (length tmp)))
		    (if (< (point) n_pt)
			(setq n_pt (+ n_pt (- (- (cdr w_bound) (car w_bound))) (length tmp))))
		    t)
		   ;; replace word or string manually
		   ((or (char-equal ?r ch_select) (char-equal ?R ch_select))
		    (setq t_bound
			  (if (char-equal ?r ch_select) w_bound s_bound))
		    (goto-char (cdr t_bound))
		    (funcall hl_on (car t_bound) (cdr t_bound))
		    (unwind-protect
			(setq tmp (pthai-completing-read-dynamic
				   "replacement for: "
				   (buffer-substring-no-properties (car t_bound)
								   (cdr t_bound))))
		      (funcall hl_off (car t_bound) (cdr t_bound)))
		    (delete-region (car t_bound) (cdr t_bound))
		    (insert tmp)
		    (goto-char (- (point) (length tmp)))
		    (setq p2 (+ p2 (- (- (cdr t_bound) (car t_bound))) (length tmp)))
		    (if (< (point) n_pt)
			(setq n_pt (+ n_pt (- (- (cdr t_bound) (car t_bound))) (length tmp))))
		    t)
		   ;; save word or string for session
		   ((or (char-equal ?a ch_select) (char-equal ?A ch_select))
		    (setq t_bound
			  (if (char-equal ?a ch_select) w_bound s_bound))
		    (setq tmp (buffer-substring-no-properties (car t_bound) (cdr t_bound)))
		    (funcall hl_off (car t_bound) (cdr t_bound))
		    (goto-char (cdr t_bound))
		    (unless (pthai-dictionary-add-word tmp)
		      (error "could not add word (%s)" tmp))
		    t)
		   ;; regenerate dictionary, goto beginning of current string
		   ((char-equal ?d ch_select)
		    (message nil)
		    (redisplay)
		    (pthai-dictionary-read-files)
		    (goto-char (car s_bound))
		    t)
		   ;; use other word splittings
		   ((char-equal split_key ch_select)
		    (setq s_idx_change t)
		    (funcall hl_off (car s_bound) (cdr s_bound))
		    (goto-char (car s_bound))
		    nil)
		   ;; exit, restore cursor position
		   ((char-equal ?x ch_select)
		    (funcall hl_off (car s_bound) (cdr s_bound))
		    (setq c_quit t))
		   ;; exit, leave cursor at point
		   ((char-equal ?X ch_select)
		    (funcall hl_off (car s_bound) (cdr s_bound))
		    (setq n_pt (car w_bound)
			  c_quit t))
		   ;; invalid selection
		   (t
		    nil)))
	    (message nil)
	    (redisplay)))))
    (progress-reporter-done p_reporter)
    (goto-char n_pt)
    (if c_quit
	(message "exit."))))

(defun pthai-spell-buffer()
  "spell check thai words in buffer"
  (interactive)
  (pthai-spell-region (point-min) (point-max)))

(defun pthai-spell-buffer-from-point()
  "spell check thai words in from point to end of buffer"
  (interactive)
  (pthai-spell-region (point) (point-max)))

(defun pthai-spell-line()
  "spell check thai words on current line"
  (interactive)
  (pthai-spell-region (line-beginning-position) (line-end-position)))

(defun pthai-spell-string(&optional aword)
  "spell check one or more contiguous thai words (a string) at point.
return nil if point not at thai string.  with option aword,
don't do word splitting on string"
  (interactive)
  (let ((bounds (pthai-bounds-of-thaistring-at-point)))
    (if (not bounds)
	nil
      (pthai-spell-region (car bounds) (cdr bounds) aword))))

(defun pthai-spell-word()
  "spell check thai string as a single word at point without any
word splitting. return nil if point not at thai string"
  (interactive)
  (pthai-spell-string t))

(defun pthai-complete-word()
  "try to complete and insert thai word at point, similar to
'ispell-complete-word. if no word selected return nil, else
return selected word"
  (interactive)
  (let* ((bound (pthai-bounds-of-thaistring-at-point))
	 (word (and bound
		    (buffer-substring-no-properties (car bound)
						    (cdr bound))))
	 tmp)
    (setq tmp (pthai-completing-read-dynamic "word: " word))
    (when (> (length tmp) 0)
      (if bound
	  (delete-region (car bound) (cdr bound)))
      (insert tmp))
    (if (> (length tmp) 0)
	tmp nil)))

(defun pthai-dictionary-parse-filelist(files)
  "take a list of files and directories, return full paths of all
non-dot files and files in directories"
  (let (tmp dfiles rfiles)
    (while files
      (setq files (mapcar 'expand-file-name files)
	    files (sort (seq-uniq files) 'string-lessp)
	    tmp (car files)
	    files (cdr files))
      (when tmp
	(cond
	 ;; skip .dot files/dirs
	 ((string-match-p "^\\." (file-name-nondirectory tmp))
	  nil)
	 ;; save list of files in dirs, skipping dot files and subdirs
	 ((file-directory-p tmp)
	  (setq dfiles (directory-files tmp nil "^\\([^.]\\|\\.[^.]\\|\\.\\..\\)")
		dfiles (mapcar (lambda (x) (concat tmp "/" x)) dfiles)
		dfiles (seq-filter (lambda (x) (not (file-directory-p x))) dfiles)
		files (append files dfiles)))
	 ;; normal file
	 ((file-exists-p tmp)
	  (setq rfiles (cons tmp rfiles))))))
    (sort (seq-uniq rfiles) 'string-lessp)))

(defun pthai-dictionary-add-word(thaiword &optional def english_class thai_class)
  "update pthai-dictionary with new entry or merge with existing
entry. thaiword is a thai string, other args can be nil, string,
or a list of strings. returns merged definition of word that was
inserted"
  (interactive)
  (if (< (length thaiword) 1)
      nil
    (let (ilist tmp)
      (unless (string-match-p "^\\ct+$" thaiword)
	(error "invalid thai word (%s)" thaiword))
      ;; set up lists if thaiword has any definition, else just add the
      ;; thaiword without any defs
      (if (or def english_class thai_class)
	  (setq ilist (mapcar (lambda(x) (if (listp x) x (list x)))
			      (list def english_class thai_class)))
	(setq ilist '(nil nil nil)))
      ;; merge if entry already exists
      (if (setq tmp (gethash thaiword pthai-dictionary))
	  (setq ilist (seq-mapn #'append ilist tmp)))
      ;; delete duplicates, nil
      (setq ilist
	    (mapcar (lambda(x) (sort (delq nil (delete-dups x)) 'string-lessp))
		    ilist))
      (puthash thaiword ilist pthai-dictionary))))

(defun pthai-dictionary-remove-word(thaiword)
  "remove word from dictionary, return t if word in dictionary,
  nil otherwise"
  (interactive "sthai word to remove: ")
  (cond
   ((gethash thaiword pthai-dictionary)
    (remhash thaiword pthai-dictionary)
    t)
   (t
    nil)))

(defun pthai-dictionary-find-regexp(regexp)
  "display list of thai words from dictionary matching thai regexp"
  (interactive "sthai regexp: \n")
  (let ((bname "*pthai-dictionary-find-regexp*"))
    (set-buffer (get-buffer-create bname))
    (erase-buffer)
    (dolist (tmp (hash-table-keys pthai-dictionary))
      (if (string-match-p regexp tmp)
	  (insert tmp "\n")))
    (sort-lines nil (point-min) (point-max))

    (if (= (point) 1)
	(progn
	  (kill-buffer bname)
	  (message "no matches found for %s" regexp))
      (switch-to-buffer bname)
      (goto-char (point-min)))))

(defun pthai-dictionary-grep(word)
  "for any word, find dictionary files names and line numbers
containing word, output to a grep-mode buffer"
  (interactive (list (pthai-completing-read-dynamic "text: " nil)))
  (let ((coding-system-for-read 'utf-8)
	(coding-system-for-write 'utf-8)
	(buffer-file-coding-system 'utf-8)
	(files (pthai-dictionary-parse-filelist pthai-dictionary-list))
	(w_rexp (concat "\\b" word "\\b"))
	(bufname "*pthai-dictionary-grep*")
	buf rlist)
    (setq files (cons pthai-pkg-wordlist files))
    (dolist (tmp files)
      (with-temp-buffer
	(insert-file tmp)
	(goto-char (point-min))
	(while (re-search-forward w_rexp nil t)
	  (setq rlist
		(cons (list tmp
			    (line-number-at-pos)
			    (buffer-substring (line-beginning-position)
					      (line-end-position)))
		      rlist)))))
    (if (< (length rlist) 1)
	(message "no files found containing %s" word)
      (setq rlist
	    (sort (seq-uniq rlist)
		  (lambda(a b)
		    (if (string= (nth 0 a) (nth 0 b))
			(< (nth 1 a) (nth 1 b))
		      (string< (nth 0 a) (nth 0 b))))))
      (setq rlist (mapcar
		   (lambda(x)
		     (concat (nth 0 x) ":"
			     (number-to-string (nth 1 x)) ":"
			     (nth 2 x)))
		   rlist))
      (setq buf (get-buffer-create bufname))
      (switch-to-buffer buf)
      (read-only-mode 0)
      (erase-buffer)
      (insert "pthai-dictionary-grep for " word "\n\n")
      (insert (string-join rlist "\n") "\n")
      (goto-char (point-min))
      (grep-mode))))

(defun pthai-dictionary-parse-classifiers(str)
  "helper function to parse definition line and return list like
((a b c) (ก ด ห))"
  (let ((pos 0)
	eclass tclass tmp)
    (while (string-match "\\({.*?}\\)" str pos)
      (setq tmp (match-string 1 str))
      (setq pos (match-end 1))
      (if (string-match-p "\\ct+" tmp)
	  (setq tclass (append (split-string tmp " *[{} \t]+ *" t) tclass))
	(setq eclass (append (split-string tmp " *[{},\t]+ *" t) eclass))))
    (list eclass tclass)))

(defun pthai-dictionary-read-region(r1 r2 &optional clear_all read_twt)
  "parse selected region into pthai-dictionary. if optional
clear_all arg, reset dictionary first.  if read_twt arg, re-read
the 'thai-word-table.  returns number of words read.  assumes
file is in utf-8 format and lines contain a single thai word at
beginning of the line, or a thai \"definition\" line"
  (interactive "r\nP")
  (let* ((coding-system-for-read 'utf-8)
	 (coding-system-for-write 'utf-8)
	 (buffer-file-coding-system 'utf-8)
	 (cbuf (current-buffer))
	 (count 0)
	 tword def_line class_info)
    (if clear_all
      (clrhash pthai-dictionary))
    (if read_twt
	(pthai-twt-read))
    (with-temp-buffer
      (insert-buffer-substring-no-properties cbuf r1 r2)
      ;; remove # comments
      (goto-char (point-min))
      (while (re-search-forward "#.*$" nil t)
	(replace-match ""))
      ;; remove stuff in parentheses
      (goto-char (point-min))
      (while (re-search-forward "[ \t]*(.*?)[ \t]*" nil t)
	(replace-match " "))
      ;; consolidate spaces
      (goto-char (point-min))
      (while (re-search-forward "[ \t]+" nil t)
	(replace-match " "))
      ;; remove double quote marks
      (goto-char (point-min))
      (while (re-search-forward "\"" nil t)
	(replace-match ""))
      ;; convert all to lowercase
      (downcase-region (point-min) (point-max))
      ;; read words without definitions
      (goto-char (point-min))
      (while (re-search-forward "^\\(\\ct+\\)$" nil t)
	(setq count (+ 1 count))
	(pthai-dictionary-add-word (match-string 1)))
      ;; read words with "dictionary" format
      (goto-char (point-min))
      (while (re-search-forward "^\\(\\ct+\\) \\(.*?\\) *$" nil t)
       	(when (and (setq tword (match-string 1))
       		   (setq def_line (match-string 2))
       		   (> (length def_line) 0))
	  (cond
	   ;; parse classifier info and strip from definition
	   ((string-match-p "[{}]" def_line)
	    (setq class_info (pthai-dictionary-parse-classifiers def_line))
	    (setq def_line (replace-regexp-in-string " ?{.*?} ?" "" def_line)))
	   (t
	    (setq class_info nil)))
	  ;; save thai word, definition, classifier info	  
	  (pthai-dictionary-add-word tword
				     (split-string def_line " ?, ?" t)
				     (nth 0 class_info)
				     (nth 1 class_info))
	  (setq count (+ 1 count)))))
    (if pthai-verbose-wordloads
	(message "%d words read" count))
    count))

(defun pthai-dictionary-read-line()
  "parse current line into pthai-dictionary. returns number of
words read"
  (interactive)
  (pthai-dictionary-read-region (line-beginning-position)
				(line-end-position)))

(defun pthai-dictionary-read-buffer(&optional clear_all read_twt)
  "parse current buffer into pthai-dictionary.  see
'pthai-dictionary-read-region for arg info. returns number of
words read"
  (interactive)
  (pthai-dictionary-read-region
   (point-min) (point-max) clear_all read_twt))

(defun pthai-dictionary-read-file(fname &optional clear_all read_twt)
  "parse file into into pthai-dictionary. see
'pthai-dictionary-read-region for arg info. returns number of
words read"
  (interactive "fpath to dictionary name: \n")
  (let ((coding-system-for-read 'utf-8)
	(coding-system-for-write 'utf-8)
	(buffer-file-coding-system 'utf-8))
    (unless (and (> (length fname) 0)
		 (setq fname (expand-file-name fname))
		 (file-exists-p fname)
		 (not (file-directory-p fname)))
      (error (concat "invalid dictionary \"" fname "\"")))
    (with-temp-buffer
      (insert-file-contents fname)
      (if pthai-verbose-wordloads
	  (message "reading %s..." fname))
      (pthai-dictionary-read-buffer clear_all read_twt))))

(defcustom pthai-dictionary-list
  (list (concat pthai-default-directory "dictionary"))
  "list of files and directories that contain thai dictionary files"
  :group 'pthai
  :set (lambda (sym val)
	 (set-default sym val)
	 (let (tmp)
	   (dolist (tmp (pthai-dictionary-parse-filelist val))
	     (if (file-exists-p tmp)
		 (pthai-dictionary-read-file tmp)))
	   (unless pthai-bootstrap
	     (when pthai-verbose-wordloads
	       (setq tmp (hash-table-count pthai-dictionary))
	       (message "%d words in pthai-dictionary" tmp)))))
  :type '(repeat file))

(defun pthai-dictionary-read-files(&optional clear)
  "read dictionary files defined in pthai-dictionary-list.  if
optional clear, clear existing entries.  returns total number of
words in dictionary"
  (interactive "P")
  (let (tmp)
    (when clear
      (clrhash pthai-dictionary)
      (pthai-twt-read))
    (mapcar 'pthai-dictionary-read-file
	    (pthai-dictionary-parse-filelist pthai-dictionary-list))
    (when pthai-verbose-wordloads
      (setq tmp (hash-table-count pthai-dictionary))
      (message "%d unique words in pthai-dictionary" tmp))
    tmp))

(defun pthai-dictionary-save(filename &optional nodefs)
  "save dictionary words to filename.  if option nodefs given,
only save words without any definitions"
  (interactive "FName of file to save to: ")
  (let* ((coding-system-for-read 'utf-8)
	 (coding-system-for-write 'utf-8)
	 (buffer-file-coding-system 'utf-8)
	 (words (sort (hash-table-keys pthai-dictionary) 'string-lessp))
	 tlist olist)
    (with-temp-buffer
      (if nodefs
	  (dolist (tmp words)
	    (insert tmp "\n"))
	(dolist (tmp words)
	  (if (pthai-dictionary-undef-p tmp)
	      (insert tmp "\n")
	    (setq olist nil)
	    (setq tlist (gethash tmp pthai-dictionary))
	    (when (nth 2 tlist)
	      (setq olist (cons (concat " {" (string-join (nth 2 tlist) " ") "}") olist)))
	    (when (nth 1 tlist)
	      (setq olist (cons (concat " {" (string-join (nth 1 tlist) ", ") "}") olist)))
	    (when (nth 0 tlist)
	      (setq olist (cons (concat (string-join (nth 0 tlist) ", ")) olist)))
	    (insert tmp "\t" (string-join olist ", ") "\n"))))
      (write-region nil nil filename))
    (length (hash-table-keys pthai-dictionary))))

(defun pthai-reverse-lookup(word ltype)
  "helper function for reverse look ups:
  word                 ltype      returns (list of ... or nil if not found)       example
  ====                 =====      ================                                =======
  english word         'def       thai definitions                                dog => (สุนัข หมา)
  english classifier   'eclass    thai words that are classified by english word  laws => (กฎ)
  thai classifier      'tclass    thai words that are classified by thai word     วงกบ => (กรอบ วง)"
  (let* (tmp vals rlist
	     (ltypes (list 'def 'eclass 'tclass))
	     (n (seq-position ltypes ltype)))
    (unless n
      (error "invalid ltype: %s" ltype))
    (dolist (tmp (hash-table-keys pthai-dictionary))
	(setq vals (nth n (gethash tmp pthai-dictionary)))
	(if (member word vals)
	    (setq rlist (append rlist (list tmp)))))
    (setq rlist (sort (delete-dups rlist) 'string-lessp))))

(defun pthai-dictionary-p(thaiword)
  "return non-nil if thaiword in dictionary, regardless of
whether it has a definition or not"
  (if (gethash thaiword pthai-dictionary)
      t nil))

(defun pthai-dictionary-def-p(thaiword)
  "return non-nil if thaiword in dictionary with a definition or
has classifier info"
  (let ((tmp (gethash thaiword pthai-dictionary)))
    (if (and tmp
	     (or (nth 0 tmp)
		 (nth 1 tmp)
		 (nth 2 tmp)))
      t nil)))

(defun pthai-dictionary-undef-p(thaiword)
  "return non-nil if thaiword in dictionary and is without any
definition. returns nil if not in dictionary or in dictionary and
has any definition"
  (let (tmp)
    (if (and (setq tmp (gethash thaiword pthai-dictionary))
	     (not (nth 0 tmp))
	     (not (nth 1 tmp))
	     (not (nth 2 tmp)))
	t nil)))

(defun pthai-dictionary-word-count()
  "return number of words in dictionary.  message user when
called interactively, displaying counts of words with and without
definitions"
  (interactive)
  (let ((total (length (hash-table-keys pthai-dictionary)))
	(defined 0)
	(undefined 0))
    (dolist (tmp (hash-table-keys pthai-dictionary))
      (if (pthai-dictionary-undef-p tmp)
	  (setq undefined (+ 1 undefined))
	(setq defined (+ 1 defined))))
    (when (called-interactively-p 'any)
      (message
       "defined/undefined/total = %d/%d/%d"
       defined undefined total))
    total))

(defun pthai-classifier-p(thaiword)
  "return non-nil if thaiword is a classifier in dictionary"
  (let (tmp)
    (if (and (setq tmp (gethash thaiword pthai-dictionary))
	     (or (nth 1 tmp) (nth 2 tmp)))
	t nil)))

(defun pthai-lookup(word &optional force_message)
  "look up definition of thai or english word. returns list of
english or thai words.  if called interactively or forced,
message user definition, '*' if word in dictionary but not
defined, or '?' if word not found in any dictionary"
  (interactive (list (pthai-completing-read-dynamic "word: " nil)))
  (let (wlist)
    (setq word (downcase word))
    (if (string-match-p "\\ct+" word)
	(setq wlist (nth 0 (gethash word pthai-dictionary)))
      (setq wlist (pthai-reverse-lookup word 'def)))
    (if (and (not wlist)
	     (string-match-p pthai-regexp-any-number word))
	(setq wlist (list (pthai-number word))))
    (when (or (called-interactively-p 'any) force_message)
      (cond
       ;; defined in dictionary
       (wlist
     	(message "%s = %s" word (string-join wlist ", ")))
       ;; in dictionary but undefined
       ((pthai-dictionary-p word)
	(message "%s = *" word))
       ;; unknown word
       (t
	(message "%s = ?" word))))
    wlist))

(defun pthai-lookup-at-point()
  "same as pthai-lookup but do word at point"
  (interactive)
  (let ((bounds (or (pthai-bounds-of-thaiword-at-point)
		    (bounds-of-thing-at-point 'word))))
    (unless bounds
      (error "could not determine bounds"))
    (pthai-highlight-say-bounds bounds nil t)))

(defun pthai-lookup-and-insert(word)
  "look up a word (english or thai) and insert meaning at point.
if more than one definition found, select from choices found"
  (interactive (list (pthai-completing-read-dynamic "word: " nil)))
  (let (tlist tlen)
    (save-excursion
      (setq word (downcase word)
	    tlist (pthai-lookup word)
	    tlen (length tlist))
      (cond
       ((= 0 tlen)
	(message "%s not found in dictionary" word))
       ((= 1 tlen)
	(insert (car tlist)))
       (t
	(insert (pthai-completing-read-full "meanings: " tlist)))))))

(defun pthai-lookup-classifier-detail(word)
  "return detailed classifier information for word. see
'pthai-lookup-classifier"
  (interactive (list (pthai-completing-read-dynamic "word: " nil)))
  (pthai-lookup-classifier word t))

(defun pthai-lookup-classifier(word &optional detail force_message)
    "return classifier information for word:
word                      return value (list of ...)
====                      ============
thai word, classifier     '( (defintions) '(eng classifiers) '(thai classifiers) ) 
thai word, not classifer  '( thai classifiers for thai word )
english word              '( thai classifiers for english word )

    message user information if called interactively. if optional
detail arg given, print out complete definitions for all
classifier words (for 2nd and 3rd cases, 1st case will by default
when interactive"
  (interactive (list (pthai-completing-read-dynamic "word: " nil)))
  (let (tmp clist ret
	    (bname (concat "*pthai-classifier-info-" word "*")))
    (setq word (downcase word))
    (cond
     ;; thai word that is a classifier, e.g. ตัว
     ((pthai-classifier-p word)
      (setq ret (gethash word pthai-dictionary))
      (setq clist (list word)))
     ;; thai word that has classifiers, e.g. แมว
     ((string-match-p "\\ct+" word)
      (setq ret (pthai-reverse-lookup word 'tclass))
      (setq clist ret))
     ;; english word, e.g. dog
     ((string-match-p "^\\Ct+$" word)
      ;; look up classifiers directly and reverse lookup thai defs and
      ;; the classifiers for them
      (setq clist (pthai-reverse-lookup word 'eclass))
      (dolist (tmp (pthai-lookup word))
	(setq clist (append (pthai-reverse-lookup tmp 'tclass) clist)))
      (setq clist (seq-filter (lambda (x) x) clist))
      (setq clist (sort (delete-dups clist) 'string-lessp))
      (setq ret clist))
     (t
      (error "error parsing classifier info")))

    (when (or detail (called-interactively-p 'any))
      (with-temp-buffer
	(cond
	 ;; word is a classifier
	 ((pthai-classifier-p word)
	  (insert (pthai-classifier-info word)))
	 ;; found classifiers for word, get list of classifiers, or details
	 ;; for each classifier found
	 (clist
	  (if detail
	      (insert (mapconcat 'pthai-classifier-info clist "\n"))
	    (insert word " classified by: " (string-join clist ", "))))
	 ;; no classifier info found
	 (t
	  (insert (pthai-classifier-info word))))
	(setq tmp (buffer-string)))
      ;; if large result, open new buffer
      (if (< (length tmp) 80)
	  (message tmp)
	(set-buffer (get-buffer-create bname))
	(erase-buffer)
	(insert tmp)
	(switch-to-buffer bname)))
    ret))

(defun pthai-lookup-classifier-at-point(&optional detail)
  "lookup classifier info for thai word at point"
  (interactive)
  (let* ((bounds (or (pthai-bounds-of-thaiword-at-point)
		     (bounds-of-thing-at-point 'word)))
	 (word (and bounds
		    (buffer-substring-no-properties (car bounds)
						    (cdr bounds)))))
    (unless word
      (error "thai word not at point"))
    (if detail
	(pthai-lookup-classifier-detail word)
      (pthai-lookup-classifier word t))))

(defun pthai-lookup-classifier-detail-at-point()
  "lookup detailed classifier info for thai word at point"
  (interactive)
  (pthai-lookup-classifier-at-point t))

(defun pthai-classifier-info(word)
  "return classifer information as a string"
  (let (cinfo cistr)
    (if (not (pthai-classifier-p word))
	(setq cistr (concat "no classifier info found for " word))

      (setq cinfo (gethash word pthai-dictionary))      
      (if (nth 0 cinfo)
	  (setq cistr (concat word " = " (string-join (nth 0 cinfo) ", ") "\n"))
	(setq cistr (concat word " is a classifier\n")))
      (if (nth 1 cinfo)
	  (setq cistr (concat cistr "classifier for: "
			     (string-join (nth 1 cinfo) ", ") "\n")))
      (if (nth 2 cinfo)
	  (setq cistr (concat cistr "classifier for: "
			     (string-join (nth 2 cinfo) " ") "\n")))
      (setq cistr (string-trim-right cistr)))
    cistr))

(defun pthai-temperature-cel2fah(celsius)
  "convert celsius to fahrenheit.  message user if called
interactively"
  (interactive "ncelsius: ")
  (let ((fah (+ (* celsius (/ 9.0 5.0)) 32.0)))
    (if (called-interactively-p 'any)
	(message "%3.1f C = %3.1f F" celsius fah))
    fah))

(defun pthai-temperature-fah2cel(fahrenheit)
  "convert fahrenheit to celsius. message user if called
interactively"
  (interactive "nfahrenheit: ")
  (let ((cel (* (- fahrenheit 32.0) (/ 5.0 9.0))))
    (if (called-interactively-p 'any)
	(message "%3.1f F = %3.1f C" fahrenheit cel))
    cel))

(defun pthai-year-greg2bud(&optional year)
  "convert gregorian year (A.D.) to buddhist year (B.E.). message
user if called interactively"
  (interactive)
  (let (bud_year
	(cur_year (string-to-number (format-time-string "%Y"))))
    (unless (integerp year)
      (setq year (round (read-number "year: " cur_year))))
    (setq bud_year (+ 543 year))
    (when (called-interactively-p 'any)
      (if (>= year 0)
	  (message "%d A.D. = B.E. %d" year bud_year)
	(message "%d B.C. = B.E. %d" (* -1 year) bud_year)))
    bud_year))

(defun pthai-year-bud2greg(&optional year)
  "convert buddhist year (B.E.) to gregorian year (A.D./B.C.).
message user if called interactively"
  (interactive)
  (let (greg_year
	(cur_year (+ 543
		     (string-to-number (format-time-string "%Y")))))
    (unless (integerp year)
      (setq year (round (read-number "year: " cur_year))))
    (setq greg_year (- year 543))
    (when (called-interactively-p 'any)
      (if (>= greg_year 0)
	  (message "B.E. %d = %d A.D." year greg_year)
	(message "B.E. %d = %d B.C. " year (* -1 greg_year))))
    greg_year))

(defun pthai-show-consonants()
  "display list of thai consonants in alphabetical order"
  (interactive)
  (message (string-join pthai-consonants " ")))

(defun pthai-show-vowels()
  "display list of thai vowels"
  (interactive)
  (message (string-join pthai-vowels " ")))

(defun pthai-show-numbers()
  "display list of thai numbers"
  (interactive)
  (message (string-join pthai-numbers " ")))

(defun pthai-count-words(&optional start end)
  "count thai words in region or buffer, similar to 'count-words,
but contiguous thai strings will be split up for counting purposes"
  (interactive)
  (let* ((p1 (or start
		 (and (use-region-p) (region-beginning))
		 (point-min)))
	 (p2 (or end
		 (and (use-region-p) (region-end))
		 (point-max)))
	 (lines (count-lines p1 p2))
	 (chars (- p2 p1))
	 (words 0)
	 (p_reporter
	  (make-progress-reporter "Counting..." p1 p2 nil 5 0.5))
	 tmp)
    (save-excursion
      (save-restriction
	(narrow-to-region p1 p2)
	(goto-char (point-min))
	(while (forward-word-strictly 1)
	  (progress-reporter-update p_reporter (point))
	  (setq words
		(+ words
		   (if (setq tmp
			     (pthai-bounds-of-thaiwords-at-point (1- (point))))
		       (length tmp)
		     1))))
	(progress-reporter-done p_reporter)))
    (when (called-interactively-p 'any)
      (message "%s has %d lines, %d words, and %d characters."
	       (if (use-region-p) "Region" "Buffer")
	       lines words chars))
    words))

(defun pthai-practice-words(wordlist)
  "practice list of thai vocab words.  list of words must already
have been read into 'pthai-dictionary"
  (let* ((prompt "[h/q/r/s/x]? ")
	 (help_line "h)elp q)uit r)epeat audio s)core x)exit")
	 right wrong select done word words
	 (p_score
	  (lambda()
	    (let* ((r_num (length right))
		   (w_num (length wrong))
		   (guesses (+ r_num w_num)))
	      (message "%d of %d correct %%%1.0f (total %d)"
		       r_num
		       guesses
		       (* 100.0 (/ (float r_num) (float (max 1 guesses))))
		       (+ guesses (length words)))))))
    (save-excursion
      (unless (> (length wordlist) 0)
	(error "no words to practice"))
      (while (not done)
	;; xxx classifiers
	(setq words (pthai-randomize (seq-filter 'pthai-lookup (or words wordlist)))
	      right nil
	      wrong nil)
	(while (and words (not done))
	  (setq word (car words))
	  (pthai-download-and-play word)
	  (setq select (downcase (read-string (concat word " " prompt))))
	  (cond
	   ;; correct word guess
	   ((seq-contains-p (pthai-lookup word) select)
	    (setq right (cons word right)
		  words (cdr words))
	    (message "correct")
	    (sit-for 1))
	   ;; help message
	   ((or (string= select "h") (string= select "?"))
	    (message help_line)
	    (sit-for 2))
	   ;; exit/quit
	   ((or (string= select "x") (string= select "q"))
	    (setq done t))
	   ;; repeat
	   ((string= select "r")
	    nil)
	   ;; score
	   ((string= select "s")
	    (funcall p_score)
	    (sit-for 1))
	   ;; wrong word guess
	   (t
	    (setq wrong (cons word wrong)
		  words (cdr words))
	    (message "wrong %s = %s" word
		     (string-join (pthai-lookup word) ", "))
	    (sit-for 1))))
	;; display current score
	(when (or right wrong)
	  (funcall p_score)
	  (sit-for 1.5))
	;; keep track of untested/wrong words. unless quitting,
	;; prompt user to continue
	(setq words (append words wrong))
	(unless (string= select "q")
	  (setq select (downcase (read-string "c)ontinue r)eset q)uit? ")))
	  (cond
	   ;; continue
	   ((string= select "c")
	    (setq done nil))
	   ;; reset
	   ((string= select "r")
	    (setq words wordlist
		  done nil))
	   ;; default/quit
	   (t
	    (setq done t))))))))

(defun pthai-practice-extract-words(&optional start end)
  "read region of thai words into pthai-dictionary.  region needs
to be in thai dictionary format. returns list of thai words read"
  (let ((p1 (or start
		(and (use-region-p) (region-beginning))
		(point-min)))
	(p2 (or end
		(and (use-region-p) (region-end))
		(point-max)))
	words)
    (save-excursion
      (save-restriction
	(narrow-to-region p1 p2)
	(pthai-dictionary-read-region (point-min) (point-max))
	(goto-char (point-min))
	(while (re-search-forward "^\\(\\ct+\\)[ \f\t\n\r\v]" nil t)
	  (setq words (cons (match-string 1) words)))))
    ;; only remove dups for small sets
    (if (> 1000 (length words))
	(setq words (seq-uniq words)))))

(defun pthai-practice-region(p1 p2)
  "practice words in highlighted region.  region needs to be in
thai dictionary format"
  (interactive "r")
  (pthai-practice-words (pthai-practice-extract-words p1 p2)))

(defun pthai-practice-buffer()
  "practice words in current buffer.  buffer needs to be in thai
dictionary format"
  (interactive)
  (pthai-practice-region (point-min) (point-max)))

(defun pthai-practice-file(file)
  "practice words from file in thai dictionary format"
  (interactive "ffile: ")
  (let* ((coding-system-for-read 'utf-8)
	 (coding-system-for-write 'utf-8)
	 (buffer-file-coding-system 'utf-8))
    (with-temp-buffer
      (insert-file-contents file)
      (pthai-practice-buffer))))

(defun pthai-unknowns(&optional start end)
  "return list of unknown thai words in a buffer or region.
unknown meaning words not in dictionary or words without
definitions in the dictionary. contiguous thai words will be
split for counting purposes.  message user when called
interactively"
  (interactive)
  (let* ((p1 (or start
		 (and (use-region-p) (region-beginning))
		 (point-min)))
	 (p2 (or end
		 (and (use-region-p) (region-end))
		 (point-max)))
	 (p_reporter
	  (make-progress-reporter "Counting..." p1 p2 nil 5 0.5))
	 words)
    (save-excursion
      (save-restriction
	(narrow-to-region p1 p2)
	(goto-char (point-min))
	(while (re-search-forward pthai-regexp-spelling-chars nil t)
	  (dolist (word (split-string (pthai-split-best (match-string 1))))
	    (progress-reporter-update p_reporter (point))
	    (when (or (not (pthai-dictionary-p word))
		      (pthai-dictionary-undef-p word))
	      (unless (or (seq-contains-p words word)
			  (string-match-p pthai-regexp-single-consonant word))
		(setq words (cons word words))))))
	(progress-reporter-done p_reporter)))
    (setq words (nreverse words))
    (when (called-interactively-p 'any)
      (if (> 1 (length words))
	  (message "no unknown words found")
	(message "%d unknown words: %s" (length words) (string-join words " "))))
    words))

(defun pthai-unknowns-count()
  "return number of unknown thai words in buffer"
  (interactive)
  (let ((len (length (pthai-unknowns))))
    (when (called-interactively-p 'any)
      (message "%d unknown words" len))
    len))

(defun pthai-unknowns-insert(ubuf)
  "insert unknown thai words from a buffer into the current
buffer, in dictionary format.  unknown words that aren't in
dictionary are prefixed with '#"
  (interactive "bbuffer to extract unknowns from: ")
  (let ((cbuf (current-buffer))
	words)
    (save-excursion
      (set-buffer ubuf)
      (setq words (pthai-unknowns nil nil))
      (set-buffer cbuf)
      (if (< (length words) 1)
	  (message "no unknown words found")
	(message "%d words inserted" (length words))
	(insert
	 (mapconcat
	  (lambda(x) (if (pthai-dictionary-p x) x (concat "#" x )))
	  words "\n"))))))

(defun pthai-x-rates-currencies()
 "from www.x-rates.com, get an alist of supported currencies and their names
'((\"USD\" . \"US Dollar\") (\"THB\" . \"Thai Baht\") ... )"
  (let ((x_url "https://www.x-rates.com")
	clist
	;; match a string like
	;;<li><a href='https://www.x-rates.com/table/?from=ARS' onclick="submitConverterArgs(this)" rel='ratestable'>Argentine Peso</a></li>
	(currency_rexp
	 (concat "<li><a href='https://www.x-rates.com/table/\\?from=\\([A-Z]\\{3\\}\\)' "
		 "onclick=.* rel='ratestable'>\\(.*\\)</a></li>")))
    (with-current-buffer (url-retrieve-synchronously x_url)
      (goto-char (point-min))
      (while (re-search-forward currency_rexp nil t)
	(setq clist (cons (cons (match-string 1) (match-string 2)) clist))))
    (unless (> (length clist) 40)
      (error "error fetching currency types"))
    (reverse clist)))

(defun pthai-x-rates-convert(from to amount)
  "convert a currency (USD THB ...) amount to another currency,
from x-rates.com.  returns converted amount as a number.  message
user conversion info if called interactively"
  (let ((x_url "https://www.x-rates.com")
	(curr_alist (pthai-x-rates-currencies))
	from_name
	to_name
	url
	mstr
	converted)
    (unless (and amount (numberp amount) (> amount 0.0))
      (error "invalid monetary amount"))
    (setq amount (number-to-string amount))
    ;; sanity check dollar & baht, and then look up names
    (unless (and (assoc "USD" curr_alist)
		 (assoc "THB" curr_alist)
		 (setq to_name (cdr (assoc to curr_alist)))
		 (setq from_name (cdr (assoc from curr_alist))))
      (error (concat "unknown currency to/from type(s): to=" to " from=" from)))
    ;; fetch current rates, then match something like
    ;; <a href='https://www.x-rates.com/graph/?from=USD&amp;to=THB'>30.171404</a>
    (setq url (concat x_url "/table/?from=" from "&amount=" amount))
    (setq mstr (concat "?from=" from "&amp;to=" to "'>"))
    (with-current-buffer (url-retrieve-synchronously url)
      (goto-char (point-min))
      (and (search-forward mstr nil t)
	   (re-search-forward "\\([0-9\.]+\\)" nil t)
	   (setq converted (format "%3.2f" (string-to-number (match-string 1))))))
    (unless converted
      (error (concat "download conversion match not found for " url)))
    (if (not (or executing-kbd-macro noninteractive))
	(message "%s %s = %s %s" amount from_name converted to_name))
    (string-to-number converted)))

(defun pthai-currency-convert(&optional from_curr to_curr amount)
  "convert between arbitrary currencies, prompt user for any optional values"
  (interactive)
  (let* ((c_list (pthai-x-rates-currencies))
	 (c_curr (mapcar 'car c_list))
	 (c_choices (mapcar (lambda (x)
			       (cons (format "%s (%s)" (cdr x) (car x)) (car x)))
			     c_list))
	 from_desc to_desc)
    (unless (and from_curr (seq-contains c_curr from_curr))
      (setq from_curr
	    (cdr (assoc (completing-read "from currency: " c_choices nil t) c_choices))))
    (or (setq from_desc (cdr (assoc from_curr c_list)))
	(error "unknown from currency description"))
    (unless (and to_curr (seq-contains c_curr to_curr))
      (setq to_curr
	    (cdr (assoc (completing-read "to currency: " c_choices nil t) c_choices))))
    (or (setq to_desc (cdr (assoc to_curr c_list)))
	(error "unknown to currency description"))

    (while (or (not amount) (not (numberp amount)) (<= amount 0.0))
      (setq amount
	    (read-number (concat "amount of " from_desc " [" from_curr "] to convert: ") 1.0)))
    ;;(message "%d \"%s\" \"%s\" => \"%s\" \"%s\"" amount from_curr from_desc to_curr to_desc)
    (pthai-x-rates-convert from_curr to_curr amount)))

(defun pthai-currency-dollar-to-baht(&optional amount)
  "convert us dollars to baht"
  (interactive)
  (pthai-currency-convert "USD" "THB" amount))

(defun pthai-currency-baht-to-dollar(&optional amount)
  "convert baht to us dollars"
  (interactive)
  (pthai-currency-convert "THB" "USD" amount))

(define-minor-mode pthai-mode
  "Toggle pthai-mode.  Minor mode to add functionality for
working with Thai text.  Adds menu of main supported elements of
'pthai module, and key re-mappings affecting word movement and
editing of Thai text.

The key commands affected are \\[right-word], \\[left-word],
\\[kill-word], \\[backward-kill-word], and \\[transpose-words]"
  ;; initial value
  :init-value nil
  :lighter "pthai"
  :keymap
  '(([remap right-word] . forward-thaiword)
    ([remap left-word] . backward-thaiword)
    ([remap kill-word] . pthai-kill-word)
    ([remap backward-kill-word] . pthai-backward-kill-word)
    ([remap transpose-words] . pthai-transpose-words)
    ("\C-xpai"  . pthai-say)
    ("\C-xpaw"  . pthai-say-word-at-point)
    ("\C-xpas"  . pthai-say-string-at-point)
    ("\C-xpal"  . pthai-say-line)
    ("\C-xpaol" . pthai-say-line-thai-only)
    ("\C-xpab"  . pthai-say-buffer)
    ("\C-xpaob" . pthai-say-buffer-thai-only)
    ("\C-xpar"  . pthai-say-region)
    ("\C-xpaot" . pthai-say-region-thai-only)
    ("\C-xpss"  . pthai-split-string-at-point)
    ("\C-xpsl"  . pthai-split-line)
    ("\C-xpsf"  . pthai-split-line-from-point)
    ("\C-xpsr"  . pthai-split-region)
    ("\C-xpsb"  . pthai-split-buffer)
    ("\C-xpul"  . pthai-unsplit-line)
    ("\C-xpur"  . pthai-unsplit-region)
    ("\C-xpub"  . pthai-unsplit-buffer)
    ("\C-xpcw"  . pthai-spell-word)
    ("\C-xpcs"  . pthai-spell-string-at-point)
    ("\C-xpcl"  . pthai-spell-line)
    ("\C-xpcb"  . pthai-spell-buffer)
    ("\C-xpcf"  . pthai-spell-buffer-from-point)
    ("\C-xpcr"  . pthai-spell-region)
    ("\C-xplu"  . pthai-lookup)
    ("\C-xplp"  . pthai-lookup-at-point)
    ("\C-xplc"  . pthai-lookup-classifier-at-point)
    ("\C-xpld"  . pthai-lookup-classifier-detail-at-point)
    ("\C-xpli"  . pthai-lookup-and-insert)
    ("\C-xpcm"  . pthai-complete-word)
    ("\C-xpdg"  . pthai-dictionary-grep)
    ("\C-xpdr"  . pthai-dictionary-read-files)
    ("\C-xpxb"  . pthai-currency-baht-to-dollar)
    ("\C-xpxd"  . pthai-currency-dollar-to-baht)))

;; initialize defaults
(unless pthai-bootstrap
  (pthai-twt-update pthai-pkg-wordlist t)
  (pthai-twt-read t)
  (pthai-dictionary-read-file pthai-pkg-sample-dictionary))

(provide 'pthai)
